# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


async def basic_async_tests(tested_class):
    secret_key_a = await tested_class().agenerate()

    # the Curve25519 class handles pulling correct values from the
    # cryptography package
    assert (
        secret_key_a.secret_bytes
        == await secret_key_a._Curve25519.asecret_bytes(secret_key_a.secret_key)
    )
    assert (
        secret_key_a.public_bytes
        == await secret_key_a._Curve25519.apublic_bytes(secret_key_a.public_key)
    )

    context = "an invalid type was allowed to be used to extract secret bytes"
    async with aignore(TypeError, if_else=aviolation(context)):
        await secret_key_a._Curve25519.asecret_bytes(secret_key_a.public_key)

    context = "an invalid type was allowed to be used to extract secret bytes"
    async with aignore(TypeError, if_else=aviolation(context)):
        await secret_key_a._Curve25519.asecret_bytes(secret_key_a)

    context = "an invalid type was allowed to be used to extract public bytes"
    async with aignore(TypeError, if_else=aviolation(context)):
        await secret_key_a._Curve25519.apublic_bytes(secret_key_a)

    # Testing equality of async constructors
    key_a_from_secret_object = await tested_class().aimport_secret_key(secret_key_a.secret_key)
    key_a_from_secret_bytes = await tested_class().aimport_secret_key(secret_key_a.secret_bytes)
    key_a_from_secret_hex = await tested_class().aimport_secret_key(secret_key_a.secret_bytes.hex())

    key_a_from_public_object = await tested_class().aimport_public_key(secret_key_a.public_key)
    key_a_from_public_bytes = await tested_class().aimport_public_key(secret_key_a.public_bytes)
    key_a_from_public_hex = await tested_class().aimport_public_key(secret_key_a.public_bytes.hex())

    context = "a falsey value secret key import didn't fail!"
    for falsey_value in (None, b"", ""):
        async with aignore(ValueError, if_else=aviolation(context)):
            await tested_class().aimport_secret_key(falsey_value)

    context = "an invalid length secret key import didn't fail!"
    for invalid_length in (1, 16, 31, 33, 48, 64):
        async with aignore(ValueError, if_else=aviolation(context)):
            await tested_class().aimport_secret_key(token_bytes(invalid_length))

    context = "a falsey value public key import didn't fail!"
    for falsey_value in (None, b"", ""):
        async with aignore(ValueError, if_else=aviolation(context)):
            await tested_class().aimport_public_key(falsey_value)

    context = "an invalid length public key import didn't fail!"
    for invalid_length in (1, 16, 31, 33, 48, 64):
        async with aignore(ValueError, if_else=aviolation(context)):
            await tested_class().aimport_public_key(token_bytes(invalid_length))

    assert len(secret_key_a.public_bytes) == 32
    assert len(secret_key_a.secret_bytes) == 32
    assert type(key_a_from_public_object.public_key) == type(key_a_from_secret_object.public_key)
    assert type(key_a_from_public_object.public_bytes) is bytes
    assert type(key_a_from_public_bytes.public_bytes) is bytes
    assert type(key_a_from_public_hex.public_bytes) is bytes
    assert key_a_from_public_object.public_bytes == key_a_from_secret_object.public_bytes
    assert key_a_from_public_bytes.public_bytes == key_a_from_secret_bytes.public_bytes
    assert key_a_from_public_hex.public_bytes == key_a_from_secret_hex.public_bytes
    assert secret_key_a.secret_bytes == key_a_from_secret_object.secret_bytes
    assert secret_key_a.secret_bytes == key_a_from_secret_bytes.secret_bytes
    assert secret_key_a.secret_bytes == key_a_from_secret_hex.secret_bytes
    assert secret_key_a.public_bytes == key_a_from_secret_object.public_bytes
    assert secret_key_a.public_bytes == key_a_from_secret_bytes.public_bytes
    assert secret_key_a.public_bytes == key_a_from_secret_hex.public_bytes
    assert secret_key_a.secret_bytes != key_a_from_secret_object.public_bytes
    assert not key_a_from_public_object.has_secret_key()
    assert not key_a_from_public_bytes.has_secret_key()
    assert not key_a_from_public_hex.has_secret_key()
    assert key_a_from_public_object.has_public_key()
    assert key_a_from_public_bytes.has_public_key()
    assert key_a_from_public_hex.has_public_key()
    assert secret_key_a.has_secret_key()
    assert secret_key_a.has_public_key()

    context = (
        "Async public key import was allowed when an instance was "
        "already initialized with a key!"
    )
    async with aignore(PermissionError, if_else=aviolation(context)) as relay:
        await secret_key_a.aimport_public_key(secret_key_a.public_bytes)
    assert "is already set" in relay.error.args[0]

    context = (
        "Async secret key import was allowed when an instance was "
        "already initialized with a key!"
    )
    async with aignore(PermissionError, if_else=aviolation(context)) as relay:
        await secret_key_a.aimport_secret_key(secret_key_a.secret_bytes)
    assert "is already set" in relay.error.args[0]


def basic_sync_tests(tested_class):
    secret_key_b = tested_class().generate()

    # the Curve25519 class handles pulling correct values from the
    # cryptography package
    assert (
        secret_key_b.secret_bytes
        == secret_key_b._Curve25519.secret_bytes(secret_key_b.secret_key)
    )
    assert (
        secret_key_b.public_bytes
        == secret_key_b._Curve25519.public_bytes(secret_key_b.public_key)
    )

    context = "an invalid type was allowed to be used to extract secret bytes"
    with ignore(TypeError, if_else=violation(context)):
        secret_key_b._Curve25519.secret_bytes(secret_key_b.public_key)

    context = "an invalid type was allowed to be used to extract secret bytes"
    with ignore(TypeError, if_else=violation(context)):
        secret_key_b._Curve25519.secret_bytes(secret_key_b)

    context = "an invalid type was allowed to be used to extract public bytes"
    with ignore(TypeError, if_else=violation(context)):
        secret_key_b._Curve25519.public_bytes(secret_key_b)

    # Testing equality of sync constructors
    key_b_from_secret_object = tested_class().import_secret_key(secret_key_b.secret_key)
    key_b_from_secret_bytes = tested_class().import_secret_key(secret_key_b.secret_bytes)
    key_b_from_secret_hex = tested_class().import_secret_key(secret_key_b.secret_bytes.hex())

    key_b_from_public_object = tested_class().import_public_key(secret_key_b.public_key)
    key_b_from_public_bytes = tested_class().import_public_key(secret_key_b.public_bytes)
    key_b_from_public_hex = tested_class().import_public_key(secret_key_b.public_bytes.hex())

    context = "a falsey value secret key import didn't fail!"
    for falsey_value in (None, b"", ""):
        with ignore(ValueError, if_else=violation(context)):
            tested_class().import_secret_key(falsey_value)

    context = "an invalid length secret key import didn't fail!"
    for invalid_length in (1, 16, 31, 33, 48, 64):
        with ignore(ValueError, if_else=violation(context)):
            tested_class().import_secret_key(token_bytes(invalid_length))

    context = "a falsey value public key import didn't fail!"
    for falsey_value in (None, b"", ""):
        with ignore(ValueError, if_else=violation(context)):
            tested_class().import_public_key(falsey_value)

    context = "an invalid length public key import didn't fail!"
    for invalid_length in (1, 16, 31, 33, 48, 64):
        with ignore(ValueError, if_else=violation(context)):
            tested_class().import_public_key(token_bytes(invalid_length))

    assert len(secret_key_b.public_bytes) == 32
    assert len(secret_key_b.secret_bytes) == 32
    assert type(key_b_from_public_object.public_key) == type(key_b_from_secret_object.public_key)
    assert type(key_b_from_public_object.public_bytes) is bytes
    assert type(key_b_from_public_bytes.public_bytes) is bytes
    assert type(key_b_from_public_hex.public_bytes) is bytes
    assert key_b_from_public_object.public_bytes == key_b_from_secret_object.public_bytes
    assert key_b_from_public_bytes.public_bytes == key_b_from_secret_bytes.public_bytes
    assert key_b_from_public_hex.public_bytes == key_b_from_secret_hex.public_bytes
    assert secret_key_b.secret_bytes == key_b_from_secret_object.secret_bytes
    assert secret_key_b.secret_bytes == key_b_from_secret_bytes.secret_bytes
    assert secret_key_b.secret_bytes == key_b_from_secret_hex.secret_bytes
    assert secret_key_b.public_bytes == key_b_from_secret_object.public_bytes
    assert secret_key_b.public_bytes == key_b_from_secret_bytes.public_bytes
    assert secret_key_b.public_bytes == key_b_from_secret_hex.public_bytes
    assert secret_key_b.secret_bytes != key_b_from_secret_object.public_bytes
    assert not key_b_from_public_object.has_secret_key()
    assert not key_b_from_public_bytes.has_secret_key()
    assert not key_b_from_public_hex.has_secret_key()
    assert key_b_from_public_object.has_public_key()
    assert key_b_from_public_bytes.has_public_key()
    assert key_b_from_public_hex.has_public_key()
    assert secret_key_b.has_secret_key()
    assert secret_key_b.has_public_key()

    with ignore(PermissionError) as relay:
        secret_key_b.import_public_key(secret_key_b.public_bytes)
    assert "is already set" in relay.error.args[0]

    with ignore(PermissionError) as relay:
        secret_key_b.import_secret_key(secret_key_b.secret_bytes)
    assert "is already set" in relay.error.args[0]


async def test_X25519(database, async_database):
    # Test class constructors
    secret_key_a = await X25519().agenerate()
    secret_key_b = X25519().generate()

    await basic_async_tests(X25519)
    basic_sync_tests(X25519)

    # exchange methods create shared keys from different instances'
    # public bytes
    shared_key_a = secret_key_a.exchange(secret_key_b.public_bytes)
    shared_key_b = secret_key_b.exchange(secret_key_a.public_bytes)
    assert bytes_are_equal(shared_key_a, shared_key_b)

    shared_key_a = await secret_key_a.aexchange(secret_key_b.public_bytes)
    shared_key_b = await secret_key_b.aexchange(secret_key_a.public_bytes)
    assert await abytes_are_equal(shared_key_a, shared_key_b)

    # exchange methods create shared keys from different instances'
    # public hex
    shared_key_a = secret_key_a.exchange(secret_key_b.public_bytes.hex())
    shared_key_b = secret_key_b.exchange(secret_key_a.public_bytes.hex())
    assert bytes_are_equal(shared_key_a, shared_key_b)

    shared_key_a = await secret_key_a.aexchange(secret_key_b.public_bytes.hex())
    shared_key_b = await secret_key_b.aexchange(secret_key_a.public_bytes.hex())
    assert await abytes_are_equal(shared_key_a, shared_key_b)

    # exchange methods create shared keys from different instances'
    # public key object
    shared_key_a = secret_key_a.exchange(secret_key_b.public_key)
    shared_key_b = secret_key_b.exchange(secret_key_a.public_key)
    assert bytes_are_equal(shared_key_a, shared_key_b)

    shared_key_a = await secret_key_a.aexchange(secret_key_b.public_key)
    shared_key_b = await secret_key_b.aexchange(secret_key_a.public_key)
    assert await abytes_are_equal(shared_key_a, shared_key_b)

    # exchange methods create shared keys from different instances'
    # secret key object
    shared_key_a = secret_key_a.exchange(secret_key_b.secret_key)
    shared_key_b = secret_key_b.exchange(secret_key_a.secret_key)
    assert bytes_are_equal(shared_key_a, shared_key_b)

    shared_key_a = await secret_key_a.aexchange(secret_key_b.secret_key)
    shared_key_b = await secret_key_b.aexchange(secret_key_a.secret_key)
    assert await abytes_are_equal(shared_key_a, shared_key_b)

    # Testing protocols
    # SYNC
    # 2DH
    peer_key = X25519().generate()
    with secret_key_b.dh2_client() as client:
        server = peer_key.dh2_server(peer_ephemeral_key=client())
        client(server.exhaust())

    assert client.result().sha3_512() == server.result().sha3_512()

    # 3DH
    with secret_key_b.dh3_client() as client:
        pkB, pkD = client()
        server = peer_key.dh3_server(peer_identity_key=pkB, peer_ephemeral_key=pkD)
        client(server.exhaust())

    assert client.result().sha3_512() == server.result().sha3_512()


    # ASYNC
    # 2DH
    peer_key = await X25519().agenerate()
    async with secret_key_b.adh2_client() as client:
        server = peer_key.adh2_server(peer_ephemeral_key=await client())
        await client(await server.aexhaust())

    assert (await client.aresult()).sha3_512() == (await server.aresult()).sha3_512()

    # 3DH
    async with secret_key_b.adh3_client() as client:
        pkB, pkD = await client()
        server = peer_key.adh3_server(peer_identity_key=pkB, peer_ephemeral_key=pkD)
        await client(await server.aexhaust())

    assert client.result().sha3_512() == server.result().sha3_512()


async def test_Ed25519(database, async_database):
    # Test class constructors
    secret_key_a = await Ed25519().agenerate()
    secret_key_b = Ed25519().generate()

    await basic_async_tests(Ed25519)
    basic_sync_tests(Ed25519)

    async_signature = await secret_key_a.asign(plaintext_bytes)
    signature = secret_key_b.sign(plaintext_bytes)

    arbitrary_verifier = Ed25519().generate()
    key_a_verifier = await Ed25519().aimport_public_key(secret_key_a.public_bytes)
    key_b_verifier = Ed25519().import_public_key(secret_key_b.public_bytes)

    # async verification succeeds when supplied a correct signature & data
    await key_a_verifier.averify(async_signature, plaintext_bytes)
    await arbitrary_verifier.averify(async_signature, plaintext_bytes, public_key=secret_key_a.public_bytes)

    # async verification succeeds when supplied an incorrect signature & data
    context = "Async verification succeeded for an invalid signature!"
    async with aignore(Ed25519.InvalidSignature, if_else=aviolation(context)):
        await key_a_verifier.averify(token_bytes(len(async_signature)), plaintext_bytes)
    context = "Async verification succeeded for an invalid signature!"
    async with aignore(Ed25519.InvalidSignature, if_else=aviolation(context)):
        await arbitrary_verifier.averify(token_bytes(len(async_signature)), plaintext_bytes, public_key=secret_key_a.public_bytes)

    # sync verification succeeds when supplied a correct signature & data
    key_b_verifier.verify(signature, plaintext_bytes)
    arbitrary_verifier.verify(signature, plaintext_bytes, public_key=secret_key_b.public_bytes)

    # sync verification succeeds when supplied an incorrect signature & data
    context = "Verification succeeded for an invalid signature!"
    with ignore(Ed25519.InvalidSignature, if_else=violation(context)):
        key_a_verifier.verify(token_bytes(len(async_signature)), plaintext_bytes)
    context = "Verification succeeded for an invalid signature!"
    with ignore(Ed25519.InvalidSignature, if_else=violation(context)):
        arbitrary_verifier.verify(token_bytes(len(async_signature)), plaintext_bytes, public_key=secret_key_a.public_bytes)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__doc__ = "Aggregated passcrypt constants."


# Config Constants:
PASSPHRASE = "passphrase"
MIN_PASSPHRASE_BYTES = 12

AAD = "aad"
DEFAULT_AAD = b""

SCHEMA = "schema"

TIMESTAMP = "timestamp"
TIMESTAMP_BYTES = 8  # measures nanoseconds
NS_TO_S_RATIO = 1_000_000_000

MB = "mb"
B_TO_MB_RATIO = 1024 * 1024
DEFAULT_MB = 64
MIN_MB = 1
MAX_MB = 256**3
MB_BYTES = 3

CPU = "cpu"
CPU_TO_DIGEST_PAIRS_PER_ROW_RATIO = 2
DEFAULT_CPU = 1
MIN_CPU = 1
MAX_CPU = 256
CPU_BYTES = 1

CORES = "cores"
DEFAULT_CORES = 4
MIN_CORES = 1
MAX_CORES = 256
CORES_BYTES = 1

SALT = "salt"
SALT_SIZE = "salt_size"
DEFAULT_SCHEMA_SALT_SIZE = 8
DEFAULT_SALT_SIZE = 16
MIN_SALT_SIZE = 4
MAX_SALT_SIZE = 256
SALT_SIZE_BYTES = 1

TAG = "tag"
TAG_SIZE = "tag_size"
DEFAULT_SCHEMA_TAG_SIZE = 16
DEFAULT_TAG_SIZE = 64
MIN_TAG_SIZE = 16

PASSCRYPT_PAD = b"\xf2"


# Dynamically Generated:
HEADER_BYTES = (
    TIMESTAMP_BYTES
    + MB_BYTES
    + CPU_BYTES
    + CORES_BYTES
    + SALT_SIZE_BYTES
)
HEADER_NIBBLES = 2 * HEADER_BYTES
HEADER_SLICE = slice(HEADER_BYTES)

TIMESTAMP_NIBBLES = 2 * TIMESTAMP_BYTES
TIMESTAMP_SLICE = slice(TIMESTAMP_BYTES)

MB_NIBBLES = 2 * MB_BYTES
MB_SLICE = slice(TIMESTAMP_BYTES, TIMESTAMP_BYTES + MB_BYTES)
MB_RESOURCE_SAFETY_RANGE = range(MIN_MB, 512)

CPU_NIBBLES = 2 * CPU_BYTES
CPU_SLICE = slice(
    TIMESTAMP_BYTES + MB_BYTES, TIMESTAMP_BYTES + MB_BYTES + CPU_BYTES
)
CPU_RESOURCE_SAFETY_RANGE = range(MIN_CPU, 33)

CORES_NIBBLES = 2 * CORES_BYTES
CORES_SLICE = slice(
    TIMESTAMP_BYTES + MB_BYTES + CPU_BYTES,
    TIMESTAMP_BYTES + MB_BYTES + CPU_BYTES + CORES_BYTES,
)
CORES_RESOURCE_SAFETY_RANGE = range(MIN_CORES, 9)

SALT_SIZE_NIBBLES = 2 * SALT_SIZE_BYTES
SALT_SIZE_SLICE = slice(
    TIMESTAMP_BYTES + MB_BYTES + CPU_BYTES + CORES_BYTES,
    HEADER_BYTES,
)

MIN_SCHEMA_BYTES = HEADER_BYTES + MIN_SALT_SIZE + MIN_TAG_SIZE


# Create a binding to the module's name for UI
passcrypt = {n: v for n, v in globals().items() if n[0].isupper()}
passcrypt["__all__"] = [*passcrypt]




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__version__ = "0.22.1"


__license__ = "AGPLv3"


__author__ = "rmlibre@riseup.net"


__PUBLIC_ED25519_KEY__ = (
    "70d1740f2a439da98243c43a4d7ef1cf993b87a75f3bb0851ae79de675af5b3b"
)


__PUBLIC_X25519_KEY__ = (
    "4457276dbcae91cc5b69f1aed4384b9eb6f933343bb44d9ed8a80e2ce438a450"
)


__doc__ = (
    "aiootp - an asynchronous pseudo one-time pad based crypto and anon"
    "ymity library."
)


from .commons import *
from .paths import *
from .asynchs import *
from .gentools import *
from .generics import *
from .randoms import *
from .ciphers import *
from .keygens import *
from .databases import *
from .__ui_coordination import *


__all__ = [
    "commons",
    *commons.__all__,
    *paths.__all__,
    "asynchs",
    *asynchs.__all__,
    "gentools",
    *gentools.__all__,
    "generics",
    *generics.__all__,
    "randoms",
    *randoms.__all__,
    "ciphers",
    *ciphers.__all__,
    "keygens",
    *keygens.__all__,
    *databases.__all__,
]


from .__engagement._report_security_issue import report_security_issue
from ._debuggers import _debuggers
from .commons import commons
from .paths import paths as _paths
from .asynchs  import asynchs
from .gentools import gentools
from .generics import generics
from .randoms import randoms
from .ciphers import ciphers
from .keygens import keygens


del paths
del databases
del __constants
del __dependencies
del __engagement
del __ui_coordination




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


async def test_xttl_functionality():
    delta = clock.delta(time_start)
    if delta < 2:
        asynchs.sleep(delta)

    # ciphertext verification fails if the hash has expired
    context = "Life-time for sync json ciphertext is malfunctioning."
    with ignore(TimeoutError, TimestampExpired, if_else=violation(context)):
        cipher.json_decrypt(test_json_ciphertext, aad=aad, ttl=1)

    context = "Life-time for async json ciphertext is malfunctioning."
    async with aignore(TimeoutError, TimestampExpired, if_else=aviolation(context)):
        await cipher.ajson_decrypt(atest_json_ciphertext, aad=aad, ttl=1)

    context = "Life-time for sync tokens is malfunctioning."
    with ignore(TimeoutError, TimestampExpired, if_else=violation(context)):
        cipher.read_token(test_token_ciphertext, aad=aad, ttl=1)

    context = "Life-time for async tokens is malfunctioning."
    async with aignore(TimeoutError, TimestampExpired, if_else=aviolation(context)):
        await cipher.aread_token(atest_token_ciphertext, aad=aad, ttl=1)


    # passcrypt verification fails if the hash has expired
    context = "Life-time for sync passcrypt hashes are malfunctioning."
    with ignore(TimeoutError, TimestampExpired, if_else=violation(context)):
        Passcrypt.verify(expired_passcrypt_hash, passphrase_0, ttl=1)

    context = "Life-time for async passcrypt hashes are malfunctioning."
    async with aignore(TimeoutError, TimestampExpired, if_else=aviolation(context)):
        await Passcrypt.averify(aexpired_passcrypt_hash, passphrase_0, ttl=1)


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


async def test_canonical_packs():
    TEST_KEYS = [token_bytes(64) for _ in range(4)]
    HASHERS = (sha3_256, sha3_512, shake_128, shake_256)

    DEFAULT_PAD = b"\x00"
    DEFAULT_BLOCKSIZE = 1
    DEFAULT_INT_BYTES = 8

    DEFAULT_PACKING = canonical_pack()
    DEFAULT_ASYNC_PACKING = await acanonical_pack()

    # the default packing is the same for async & sync
    assert DEFAULT_PACKING == DEFAULT_ASYNC_PACKING
    assert (
        DEFAULT_PACKING
        == await acanonical_pack(pad=DEFAULT_PAD, blocksize=DEFAULT_BLOCKSIZE, int_bytes=DEFAULT_INT_BYTES)
    )

    # the options can be used together
    assert (
        canonical_unpack(canonical_pack(b"", pad=b"0", blocksize=127, int_bytes=16), int_bytes=16)
        == await acanonical_unpack(await acanonical_pack(b"", pad=b"0", blocksize=127, int_bytes=16), int_bytes=16)
    )

    test_inputs = [
        [token_bytes(token_bits(8)) for _ in range(2 + token_bits(2))]
        for _ in range(4)
    ]
    test_int_bytes = [1, 2, 3, 4]
    test_blocksizes = [SHA3_256_BLOCKSIZE, SHA3_512_BLOCKSIZE, SHAKE_128_BLOCKSIZE, SHAKE_256_BLOCKSIZE]
    test_pads = [b"", b"\x01", b"\x80", b"\xff"]

    for inputs in test_inputs:
        # similar inputs do no produce the same output
        for int_bytes in test_int_bytes:
            result = canonical_pack(*inputs)
            aresult = await acanonical_pack(*inputs, int_bytes=int_bytes)
            assert result != aresult

            # but the items are still packed correctly & interoperably
            assert inputs == list(canonical_unpack(aresult, int_bytes=int_bytes))
            assert inputs == list(await acanonical_unpack(result))
        for blocksize in test_blocksizes:
            result = canonical_pack(*inputs)
            aresult = await acanonical_pack(*inputs, blocksize=blocksize)
            assert result != aresult
            assert 0 == len(aresult) % blocksize

            # but the items are still packed correctly & interoperably
            assert inputs == list(canonical_unpack(aresult))
            assert inputs == list(await acanonical_unpack(result))
        for pad, hasher, key in zip(test_pads, HASHERS, TEST_KEYS):
            result = canonical_pack(*inputs)
            aresult = await acanonical_pack(*inputs, pad=pad)
            assert result != aresult

            # but the items are still packed correctly & interoperably
            assert inputs == list(canonical_unpack(aresult))
            assert inputs == list(await acanonical_unpack(result))

            if not pad:
                continue

            # and the items & pads are used correctly in hashers
            obj = hasher()
            keyed_obj = obj.copy()
            digest_size = obj.digest_size
            encoded_key = encode_key(key, obj.block_size, pad=pad)
            packing = canonical_pack(
                (digest_size if digest_size else 64).to_bytes(DEFAULT_INT_BYTES, BIG),
                *inputs,
                blocksize=obj.block_size,
                pad=pad,
            )
            keyed_obj.update(encoded_key + packing)
            obj.update(packing)
            if digest_size:
                # keyed hashing of packed items works as expected
                assert keyed_obj.digest() == hash_bytes(*inputs, pad=pad, key=key, hasher=hasher)
                assert keyed_obj.digest() == await ahash_bytes(*inputs, pad=pad, key=key, hasher=hasher)

                # un-keyed hashing of packed items works as expected
                assert obj.digest() == hash_bytes(*inputs, pad=pad, hasher=hasher)
                assert obj.digest() == await ahash_bytes(*inputs, pad=pad, hasher=hasher)
            else:
                # keyed hashing of packed items works as expected
                assert keyed_obj.digest(64) == hash_bytes(*inputs, pad=pad, key=key, hasher=hasher, size=64)
                assert keyed_obj.digest(64) == await ahash_bytes(*inputs, pad=pad, key=key, hasher=hasher, size=64)

                # un-keyed hashing of packed items works as expected
                assert obj.digest(64) == hash_bytes(*inputs, pad=pad, hasher=hasher, size=64)
                assert obj.digest(64) == await ahash_bytes(*inputs, pad=pad, hasher=hasher, size=64)

        # the same inputs produce the same outputs
        for int_bytes in test_int_bytes:
            result = canonical_pack(*inputs, int_bytes=int_bytes)
            aresult = await acanonical_pack(*inputs, int_bytes=int_bytes)
            assert result == aresult

            # the relative location of the default pad declaration is
            # dependent on the size of integers used to represent item
            # lengths
            assert type(result) is bytes
            assert result[4 * int_bytes] == DEFAULT_PAD[0]
        for blocksize in test_blocksizes:
            result = canonical_pack(*inputs, blocksize=blocksize)
            aresult = await acanonical_pack(*inputs, blocksize=blocksize)
            assert result == aresult
            assert 0 == len(result) % blocksize
        for pad in test_pads:
            result = canonical_pack(*inputs, pad=pad)
            aresult = await acanonical_pack(*inputs, pad=pad)
            assert result == aresult

            # the default integer size of 8 bytes puts the pad item at
            # the default location
            assert (result[PACK_PAD_INDEX] == pad[0]) if pad else 1

            # the relative location of the pad declaration is dependent
            # on the size of integers used to represent item lengths
            for int_bytes in test_int_bytes:
                result = canonical_pack(*inputs, pad=pad, int_bytes=int_bytes)
                aresult = await acanonical_pack(*inputs, pad=pad, int_bytes=int_bytes)
                assert result == aresult
                assert (result[4 * int_bytes] == pad[0]) if pad else 1

    pad = b"Z"
    items = (b"testing", b"pad", b"character", b"location", b"in", b"result")
    packing = bytearray(canonical_pack(*items, pad=pad))
    assert packing[PACK_PAD_INDEX] == pad[0]


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




#
# This file is copied with minor edits from the asyncio_contextmanager
# package whose license can be found in the /licenses directory of this
# package. The edits prevent warnings from arising in the aiootp package
# when values are retrieved from async generators since they do not stop
# as the class here expects them to. For convenience the license for the
# asyncio_contextmanager package will be copied below:
#
# MIT License
#
# Copyright (c) 2017 Alexander Gorokhov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#


import functools
import inspect
import sys


__all__ = ['async_contextmanager']


class _AsyncContextManager(object):

    __slots__ = ("async_generator",)

    def __init__(self, func, args, kwargs):
        if not inspect.isasyncgenfunction(func):
            raise TypeError('%s is not async generator function' % func)
        self.async_generator = func(*args, **kwargs)

    async def __aenter__(self):
        try:
            return await self.async_generator.__anext__()
        except StopAsyncIteration as e:
            raise RuntimeError("async generator didn't yield") from None

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type == None:
            try:
                await self.async_generator.__anext__()
            except StopAsyncIteration:
                return
            else:
                raise RuntimeError("async generator didn't stop")
        else:
            if exc_val == None:
                exc_val = exc_type()
            try:
                await self.async_generator.athrow(exc_type, exc_val, exc_tb)
                return True
            except StopAsyncIteration as exc:
                return exc is not exc_val
            except RuntimeError as exc:
                if exc is exc_val:
                    return False
                if exc.__cause__ is exc_val:
                    return False
                raise
            except:
                if sys.exc_info()[1] is not exc_val:
                    raise


def async_contextmanager(func):
    """
    @async_contextmanager decorator.

    Typical usage:

        @async_contextmanager
        async def some_async_generator(<arguments>):
            <setup>
            try:
                yield <value>
            finally:
                <cleanup>

    This makes this:

        async with some_async_generator(<arguments>) as <variable>:
            <body>

    equivalent to this:

        <setup>
        try:
            <variable> = <value>
            <body>
        finally:
            <cleanup>
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return _AsyncContextManager(func, args, kwargs)

    return wrapper




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


async def test_acsprng():
    entropy = await randoms.arandom_number_generator(entropy=test_data, freshness=1)
    assert len(entropy) == 64
    assert entropy.__class__ is bytes

    entropy = await randoms.arandom_number_generator(128, freshness=0)
    assert len(entropy) == 128
    assert entropy.__class__ is bytes

    entropy = await acsprng(b"")
    assert len(entropy) == 64
    assert entropy.__class__ is bytes

    entropy = await acsprng("test")
    assert len(entropy) == 64
    assert entropy.__class__ is bytes

    key = await agenerate_key(freshness=0)
    assert len(key) == KEY_BYTES
    assert key.__class__ is bytes

    context = f"Allowed to generate a key less than {MIN_KEY_BYTES}"
    async with aignore(ValueError, if_else=aviolation(context)):
        key = await agenerate_key(size=MIN_KEY_BYTES - 1, freshness=0)

    async def try_to_make_duplicate_readouts():
        """
        The async csprng doesn't produce duplicate outputs when run in
        a multithreaded environment.
        """
        for i in range(32):
            await arandom_sleep(0.00001)
            entropy = await acsprng()

            assert entropy not in entropy_pool

            entropy_pool.add(entropy)

    entropy_pool = set()
    await Threads.agather(
        *[try_to_make_duplicate_readouts for _ in range(32)]
    )


def test_csprng():
    entropy = randoms.random_number_generator(entropy=test_data, freshness=1)
    assert len(entropy) == 64
    assert entropy.__class__ is bytes

    entropy = randoms.random_number_generator(128, freshness=0)
    assert len(entropy) == 128
    assert entropy.__class__ is bytes

    entropy = csprng(b"")
    assert len(entropy) == 64
    assert entropy.__class__ is bytes

    entropy = csprng("test")
    assert len(entropy) == 64
    assert entropy.__class__ is bytes

    key = generate_key(freshness=0)
    assert len(key) == KEY_BYTES
    assert key.__class__ is bytes

    context = f"Allowed to generate a key less than {MIN_KEY_BYTES} bytes."
    with ignore(ValueError, if_else=violation(context)):
        key = generate_key(size=MIN_KEY_BYTES - 1, freshness=0)

    def try_to_make_duplicate_readouts():
        """
        The async csprng doesn't produce duplicate outputs when run in
        a multithreaded environment.
        """
        for i in range(32):
            random_sleep(0.00001)
            entropy = csprng()

            assert entropy not in entropy_pool

            entropy_pool.add(entropy)

    entropy_pool = set()
    Threads.gather(
        *[try_to_make_duplicate_readouts for _ in range(32)]
    )


async def test_guids_uniqueness():
    assert 16 == len(GUID().new())
    assert 16 == len(await GUID().anew())

    # simulate the monotonic nature of a timestamp to test the uniqueness
    # of the algorithm outputs for all inputs between two multiples of
    # the prime used for that size category

    def raw_guid_simulator():
        nonlocal i

        i += 1
        return i

    # lowering the default minimum size to be able to efficiently do a
    # complete search of a salt space & the impacts of each possible
    # change in salt on uniqueness of outputs
    GUID._MIN_SIZE = 1
    GUID._MIN_RAW_SIZE = 0
    GUID._MIN_SALT_SIZE = 1
    GUID._COUNTER_BYTES = 0

    # testing all the combinations is impossible, in general, but even
    # testing all two byte combinations is prohibitively slow. However,
    # these variables can be customized to run extended tests
    MIN_TEST_BYTES = 1
    MAX_TEST_BYTES = 1

    # change to `True` if a random location within each search space is
    # desired
    start_at_random_location: bool = False

    previous_salts = []

    # these values do not change in the tests, but are displayed to
    # visually declare where their byte values are located in relation
    # to the raw guid which contains the timestamp & random bytes
    node_number = 0
    counter = 0

    for size in range(MIN_TEST_BYTES, MAX_TEST_BYTES + 1):
        start = token_bits(8 * size) if start_at_random_location else 0
        for salt_test in bytes_range.root(start, 256**size, size=size):
            i = -1

            guid = GUID(salt=salt_test, size=size)
            _size, gen, prime, subprime = guid._session_configuration
            assert size == _size

            isalt, osalt, xsalt = guid._encode_salt(guid._salt, prime, size)
            assert previous_salts != [isalt, osalt, xsalt]
            assert all([isalt, osalt])

            #                    |--------------- `size`-bytes ---------------|
            inner_guid = lambda: (node_number + raw_guid_simulator() + counter)
            _key = lambda: (
                xsalt ^ ((isalt * inner_guid() + osalt) % prime)
            ).to_bytes(size, BIG)

            history = set()
            for j in range(prime):
                result = _key()
                assert result not in history, (
                    f"salt_test={salt_test}--isalt={isalt}--osalt={osalt}--"
                    f"xsalt={xsalt}--size={size}--j={j}--i={i}--result={result}"
                )
                history.add(result)

            previous_salts = [isalt, osalt, xsalt]

    GUID._MIN_SIZE = MIN_GUID_BYTES
    GUID._MIN_RAW_SIZE = MIN_RAW_GUID_BYTES
    GUID._MIN_SALT_SIZE = 16
    GUID._COUNTER_BYTES = 1


async def test_guid_unmask():
    for size in range(GUID._MIN_SIZE, GUID._MAX_SIZE + 1):
        for node_number in range(0, 256, 16):
            salt = token_bytes(max([size, 16]))
            guid_gen = GUID(size=size, node_number=node_number)
            guid = guid_gen.new()
            encoded_guid = guid_gen.new(encode=True)
            unmasked_guid = guid_gen.unmask(guid)
            decoded_unmasked_guid = guid_gen.unmask(encoded_guid, decode=True)

            # sync & async unmasking are the same
            assert unmasked_guid == await guid_gen.aunmask(guid)
            assert decoded_unmasked_guid == await guid_gen.aunmask(encoded_guid, decode=True)

            # the counter starts at 1 & increments every call to new
            assert unmasked_guid.counter == b"\x01"
            assert decoded_unmasked_guid.counter  == b"\x02"

            # the node number is correctly retrieved by unmasking
            assert unmasked_guid.node_number == node_number.to_bytes(1, BIG)
            assert unmasked_guid.node_number == decoded_unmasked_guid.node_number

            # nanosecond timestamps are retrieved correctly & are always
            # increasing
            assert ns_clock.make_timestamp() > unmasked_guid.timestamp
            assert ns_clock.make_timestamp() > decoded_unmasked_guid.timestamp

            # every new guid is always unmasked to a value which is
            # larger than all previous guids
            assert guid_gen.unmask(guid_gen.new()) > unmasked_guid
            assert (
                guid_gen.unmask(guid_gen.new(encode=True), decode=True)
                > decoded_unmasked_guid
            )


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = []


__doc__ = (
    "A small collection of ``pathlib.Path`` objects with references to "
    "potentially helpful directories."
)


import os
import aiofiles
from pathlib import Path
from hashlib import sha3_256
from secrets import token_bytes
from .__constants import *
from ._exceptions import *
from ._typing import Typing
from .asynchs import aos
from .commons import make_module
from .generics import BytesIO


def CurrentPath() -> Path:
    """
    Returns a ``pathlib.Path`` object pointing to the current working
    directory.
    """
    return Path(os.getcwd()).absolute()


def RootPath() -> Path:
    """
    Returns a ``pathlib.Path`` object pointing to this module's directory.
    """
    dir_path = os.path.dirname(os.path.realpath(__file__))
    return Path(dir_path).absolute()


def PackagePath() -> Path:
    """
    Returns a ``pathlib.Path`` object pointing to this package's top
    level directory.
    """
    return RootPath().parent


def TorPath(*, path: Typing.OptionalPathStr = None) -> Path:
    """
    Returns a ``pathlib.Path`` object pointing to an optional tor
    directory.
    """
    return (Path(path).absolute() if path else RootPath()) / "tor"


def DatabasePath(*, path: Typing.OptionalPathStr = None) -> Path:
    """
    Returns a ``pathlib.Path`` object pointing to the default directory
    encrypted databases are saved to.
    """
    return (Path(path).absolute() if path else RootPath()) / "databases"


async def adeniable_filename(key: bytes, *, size: int = 8) -> str:
    """
    Xors subsequent bytes-type, ``size`` length segments of ``key`` with
    each other to create a condensed & variably forgeable hash. This
    hash is used as a filename which is deniably attributable to a
    particular ``key``.
    """
    from .generics import axi_mix

    if size > 16 or size <= 0:
        raise Issue.value_must_be_value("size", "<= 16 and > 0")
    filename = sha3_256(await axi_mix(key, size=size)).digest()
    return await BytesIO.abytes_to_filename(filename[FILENAME_HASH_SLICE])


def deniable_filename(key: bytes, *, size: int = 8) -> str:
    """
    Xors subsequent bytes-type, ``size`` length segments of ``key`` with
    each other to create a condensed & variably forgeable hash. This
    hash is used as a filename which is deniably attributable to a
    particular ``key``.
    """
    from .generics import xi_mix

    if size > 16 or size <= 0:
        raise Issue.value_must_be_value("size", "<= 16 and > 0")
    filename = sha3_256(xi_mix(key, size=size)).digest()
    return BytesIO.bytes_to_filename(filename[FILENAME_HASH_SLICE])


async def amake_salt_file(path: Path, *, key: bytes) -> None:
    """
    Creates & populates the ``path`` file with a sensitive cryptographic
    salt used to harden user databases. If ``key`` is specified then the
    filename is deterministically derived from the key so a database can
    find its own salt. These salt files have their permissions changed
    with `os.chmod` to `0o000` after they're created.
    """
    secret = token_bytes(32)
    filename = await adeniable_filename(key)
    filepath = path / filename
    async with aiofiles.open(filepath, "wb") as f:
        await f.write(secret)
    await aos.chmod(filepath, 0o000)


def make_salt_file(path: Path, *, key: bytes) -> None:
    """
    Creates & populates the ``path`` file with a sensitive cryptographic
    salt used to harden user databases. If ``key`` is specified then the
    filename is deterministically derived from the key so a database can
    find its own salt. These salt files have their permissions changed
    with `os.chmod` to `0o000` after they're created.
    """
    secret = token_bytes(32)
    filename = deniable_filename(key)
    filepath = path / filename
    with open(filepath, "wb") as f:
        f.write(secret)
    os.chmod(filepath, 0o000)


async def afind_salt_file(path: Path, *, key: bytes) -> Path:
    """
    This returns the path of a sensitive cryptographic salt used to
    harden user databases. If ``key`` is specified, the salt filename is
    deterministically derived from the key so a database can find its
    own salt.
    """
    return path / await adeniable_filename(key)


def find_salt_file(path: Path, *, key: bytes) -> Path:
    """
    This returns the path of a sensitive cryptographic salt used to
    harden user databases. If ``key`` is specified, the salt filename is
    deterministically derived from the key so a database can find its
    own salt.
    """
    return path / deniable_filename(key)


async def aread_salt_file(filepath: Typing.PathStr) -> bytes:
    """
    This returns the sensitive cryptographic salt contained within the
    file located at ``filepath``. The file has its permissions changed
    with `os.chmod` to `0o000` after its read.
    """
    try:
        await aos.chmod(filepath, 0o700)
        async with aiofiles.open(filepath, "rb") as salt_file:
            salt = await salt_file.read()
            if len(salt) >= 32:
                return salt
            else:
                raise ValueError("The salt file is empty or corrupt!")
    finally:
        await aos.chmod(filepath, 0o000)


def read_salt_file(filepath) -> bytes:
    """
    This returns the sensitive cryptographic salt contained within the
    file located at ``filepath``. The file has its permissions changed
    with `os.chmod` to `0o000` after its read.
    """
    try:
        os.chmod(filepath, 0o700)
        with open(filepath, "rb") as salt_file:
            salt = salt_file.read()
            if len(salt) >= 32:
                return salt
            else:
                raise ValueError("The salt file is empty or corrupt!")
    finally:
        os.chmod(filepath, 0o000)


async def adelete_salt_file(filepath) -> None:
    """
    This returns the sensitive cryptographic salt contained within the
    file located at ``filepath``. The file has its permissions changed
    with `os.chmod` to `0o000` after its read.
    """
    await aos.chmod(filepath, 0o700)
    await aos.remove(filepath)


def delete_salt_file(filepath) -> None:
    """
    This returns the sensitive cryptographic salt contained within the
    file located at ``filepath``. The file has its permissions changed
    with `os.chmod` to `0o000` after its read.
    """
    os.chmod(filepath, 0o700)
    os.remove(filepath)


async def AsyncSecurePath(
    path: Typing.OptionalPathStr = None,
    *,
    key: Typing.Optional[bytes] = None,
    _admin: bool = False,
) -> Path:
    """
    This constructor returns the path for files which contain sensitive
    cryptographic salts used to harden user databases. If ``key`` is
    specified then the salt filename is deterministically derived from
    the key so a database can find its own salt. These salt files have
    their permissions changed with `os.chmod` to `0o000` before & after
    they're read from or created.
    """
    path = (Path(path).absolute() if path else DatabasePath()) / "secure"
    admin_path = path / "_admin"
    if not path.exists():
        await aos.mkdir(path)
    if not admin_path.exists():
        await aos.mkdir(admin_path)
    if not key:
        return path
    elif _admin:
        filepath = await afind_salt_file(admin_path, key=key)
        if not filepath or not filepath.exists():
            await amake_salt_file(admin_path, key=key)
    else:
        filepath = await afind_salt_file(path, key=key)
        if not filepath or not filepath.exists():
            await amake_salt_file(path, key=key)
    return filepath


def SecurePath(
    path: Typing.OptionalPathStr = None,
    *,
    key: Typing.Optional[bytes] = None,
    _admin: bool = False,
) -> Path:
    """
    This constructor returns the path for files which contain sensitive
    cryptographic salts used to harden user databases. If ``key`` is
    specified then the salt filename is deterministically derived from
    the key so a database can find its own salt. These salt files have
    their permissions changed with `os.chmod` to `0o000` before & after
    they're read from or created.
    """
    path = (Path(path).absolute() if path else DatabasePath()) / "secure"
    admin_path = path / "_admin"
    if not path.exists():
        os.mkdir(path)
    if not admin_path.exists():
        os.mkdir(admin_path)
    if not key:
        return path
    if _admin:
        filepath = find_salt_file(admin_path, key=key)
        if not filepath or not filepath.exists():
            make_salt_file(admin_path, key=key)
    else:
        filepath = find_salt_file(path, key=key)
        if not filepath or not filepath.exists():
            make_salt_file(path, key=key)
    return filepath


extras = dict(
    AsyncSecurePath=AsyncSecurePath,
    CurrentPath=CurrentPath,
    DatabasePath=DatabasePath,
    PackagePath=PackagePath,
    Path=Path,
    RootPath=RootPath,
    SecurePath=SecurePath,
    TorPath=TorPath,
    __doc__=__doc__,
    __package__=__package__,
    _afind_salt_file=afind_salt_file,
    _adelete_salt_file=adelete_salt_file,
    _amake_salt_file=amake_salt_file,
    _aread_salt_file=aread_salt_file,
    _find_salt_file=find_salt_file,
    _delete_salt_file=delete_salt_file,
    _make_salt_file=make_salt_file,
    _read_salt_file=read_salt_file,
    adeniable_filename=adeniable_filename,
    deniable_filename=deniable_filename,
)


paths = make_module("paths", mapping=extras)




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


async def test_json_conversion_functions():

    aciphertext = await cipher.abytes_encrypt(plaintext_bytes)
    ciphertext = cipher.bytes_encrypt(plaintext_bytes)

    assert aciphertext != ciphertext

    aciphertext_json = await BytesIO.aciphertext_to_json(aciphertext)
    ciphertext_json = BytesIO.ciphertext_to_json(ciphertext)

    assert aciphertext_json != ciphertext_json

    assert aciphertext == await BytesIO.ajson_to_ciphertext(aciphertext_json)
    assert ciphertext == BytesIO.json_to_ciphertext(ciphertext_json)


    aad = b"aad"
    key_bundle = KeyAADBundle(key=key, salt=salt, aad=aad, allow_dangerous_determinism=True).sync_mode()
    aaad=b"async_aad"
    akey_bundle = await KeyAADBundle(key=key, salt=salt, aad=aaad, allow_dangerous_determinism=True).async_mode()

    padded_plaintext = apadded_plaintext = await Padding.apad_plaintext(plaintext_bytes)

    shmac = StreamHMAC(key_bundle)._for_encryption()
    ashmac = StreamHMAC(akey_bundle)._for_encryption()

    encipher = bytes_encipher(data(padded_plaintext), shmac)
    aencipher = abytes_encipher(adata(apadded_plaintext), ashmac)

    ciphertext = {
        CIPHERTEXT: [BytesIO.bytes_to_base64(chunk) for chunk in encipher],
        SHMAC: shmac.finalize().hex(),
        SALT: key_bundle.salt.hex(),
        IV: key_bundle.iv.hex(),
    }
    aciphertext = {
        CIPHERTEXT: [BytesIO.bytes_to_base64(chunk) for chunk in await aunpack(aencipher).alist()],
        SHMAC: (await ashmac.afinalize()).hex(),
        SALT: akey_bundle.salt.hex(),
        IV: akey_bundle.iv.hex(),
    }

    # no two ciphertexts are ever the same
    assert ciphertext != aciphertext

    aads = (aad, aaad)
    ciphertexts = (ciphertext, aciphertext)
    for i, (_aad, json_message) in enumerate(zip(aads, ciphertexts)):
        message = BytesIO.json_to_ciphertext(json_message)
        amessage = await BytesIO.ajson_to_ciphertext(json_message)

        # decryption of manually reconstructed ciphertext from json
        # doesn't fail & is sound
        assert plaintext_bytes == cipher.bytes_decrypt(message, aad=_aad)
        assert plaintext_bytes == cipher.bytes_decrypt(amessage, aad=_aad)
        assert plaintext_bytes == await cipher.abytes_decrypt(message, aad=_aad)
        assert plaintext_bytes == await cipher.abytes_decrypt(amessage, aad=_aad)

        # use of wrong aad fails
        context = "Alteration of aad was not noticed!"
        with ignore(StreamHMAC.InvalidSHMAC, if_else=violation(context)):
            cipher.bytes_decrypt(message, aad=aads[(i + 1) % 2])

        # async use of wrong aad fails
        context = "Async alteration of aad was not noticed!"
        async with aignore(StreamHMAC.InvalidSHMAC, if_else=aviolation(context)):
            await cipher.abytes_decrypt(message, aad=aads[(i + 1) % 2])


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = []


__doc__ = (
    "A module used to aggregate commonly used constants & arbitrary oth"
    "er namespace utilities."
)


import sys
import copy
import json
import types
import asyncio
from math import ceil
from secrets import token_bytes
from os import linesep as sep
from hashlib import sha3_256, sha3_512, shake_128, shake_256
from .__constants import *
from ._typing import Typing as t


class DeletedAttribute:
    """
    Creates objects which raise the result of a callback function if the
    object is queried for attributes.
    """

    __slots__ = ("_callback",)

    def __init__(self, callback: t.Callable[..., Exception]) -> "self":
        self._callback = callback

    def __getattr__(self, name: str) -> None:
        raise self._callback()


class Slots:
    """
    A base class which allow subclasses to create very efficient
    instances, with explicitly declared attributes in their `__slots__`.
    """

    __slots__ = ()

    def __init__(self, **kwargs) -> "self":
        """
        Maps the user-defined kwargs to the instance attributes. If a
        subclass defines a `__slots__` list, then only variables with
        names in the list can be admitted to the instance. Defining
        classes with __slots__ can greatly increase memory efficiency if
        a system instantiates many objects of the class.
        """
        for name, value in kwargs.items():
            setattr(self, name, value)

    def __bool__(self) -> bool:
        """
        If the instance is empty then return False, otherwise True.
        """
        return any(self)

    def __len__(self) -> int:
        """
        Returns the number of elements in the instance.
        """
        return sum(1 for name in self.__slots__ if hasattr(self, name))

    def __dir__(self) -> t.List[str]:
        """
        Returns the list of names in the instance.
        """
        return [name for name in self.__slots__ if hasattr(self, name)]

    def __contains__(self, name: str) -> bool:
        """
        Returns a bool of ``name``'s membership in the instance.
        """
        return hasattr(self, name)

    def __setitem__(self, name: str, value: t.Any) -> None:
        """
        Transforms bracket item assignment into dotted assignment on the
        instance.
        """
        setattr(self, name, value)

    def __getitem__(self, name: str) -> t.Any:
        """
        Transforms bracket lookup into dotted access on the instance.
        """
        return getattr(self, name)

    def __delitem__(self, name: str) -> None:
        """
        Deletes the item ``name`` from the instance.
        """
        delattr(self, name)

    def __repr__(self, *, mask: bool = True) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        from ._debuggers import DebugControl

        if not mask or DebugControl.is_debugging():
            items = f',{sep}    '.join(
                f"{var}={repr(val)}"
                for var, val in self.items()
                if not str(var).startswith("_")
            )
        else:
            items = f',{sep}    '.join(
                f"{var}="
                f"{OMITTED} of {type(val)} "
                f"{(val and 'is <truthy>') or 'is <falsey>'}"
                for var, val in self.items()
                if not str(var).startswith("_")
            )
        cls = self.__class__.__qualname__
        return f"{cls}({f'{sep}    {items},{sep}' if items else ''})"

    async def __aiter__(self) -> t.AsyncGenerator[None, t.Any]:
        """
        Allows an instance to be unpacked with with async iteration.
        """
        for name in self.__slots__:
            if hasattr(self, name):
                await asyncio.sleep(0)
                yield name

    def __iter__(self) -> t.Generator[None, t.Any, None]:
        """
        Allows an instance to be unpacked with tools like ``dict`` &
        ``list``.
        """
        for name in self.__slots__:
            if hasattr(self, name):
                yield name

    def keys(self) -> t.Generator[None, t.Any, None]:
        """
        Yields the names of all items in the instance.
        """
        yield from (
            name
            for name in self.__slots__
            if hasattr(self, name)
        )

    def values(self) -> t.Generator[None, t.Any, None]:
        """
        Yields the values of all items in the instance.
        """
        yield from (
            getattr(self, name)
            for name in self.__slots__
            if hasattr(self, name)
        )

    def items(self) -> t.Generator[None, t.Tuple[t.Any, t.Any], None]:
        """
        Yields the name, value pairs of all items in the instance.
        """
        yield from (
            (name, getattr(self, name))
            for name in self.__slots__
            if hasattr(self, name)
        )


class Namespace(Slots):
    """
    A simple wrapper for turning mappings into Namespace objects that
    allow dotted lookup and assignment on those mappings. Also, provides
    utilities for copying mappings into other containers, & turning
    mappings into stand-alone, first-class modules.
    """

    __slots__ = ("__dict__",)

    def __init__(self, mapping={}, **kwargs) -> "self":
        """
        Maps the user-defined mapping & kwargs to the Namespace's
        instance dictionary.
        """
        if mapping.__class__ in {str, bytes, bytearray}:
            mapping = json.loads(mapping)
        self.__dict__.update(mapping)
        self.__dict__.update(kwargs)

    @property
    def __all__(self) -> t.Generator[None, t.Hashable, None]:
        """
        Allows users that have turned their Namespace into a Module
        object to do a ``from namespace import *`` on the contents of
        the namespace's mapping. This method excludes exporting private
        methods & attributes.
        """
        return [var for var in self.__dict__ if str(var)[0] != "_"]

    @property
    def namespace(self) -> dict:
        """
        Cleaner name for users to access the instance's dictionary.
        """
        return self.__dict__

    def __bool__(self) -> bool:
        """
        If the namespace is empty then return False, otherwise True.
        """
        return bool(self.__dict__)

    def __len__(self) -> int:
        """
        Returns the number of elements in the Namespace's mapping.
        """
        return len(self.__dict__)

    def __dir__(self) -> t.List[t.Hashable]:
        """
        Returns the list of names in the Namespace's mapping.
        """
        return [*self.__dict__]

    def __contains__(self, variable=None) -> bool:
        """
        Returns a bool of ``variable``'s membership in the instance
        dictionary.
        """
        return variable in self.__dict__

    def __setitem__(self, variable, value) -> None:
        """
        Transforms bracket item assignment into dotted assignment on the
        Namespace's mapping.
        """
        setattr(self, variable, value)

    def __getitem__(self, variable) -> t.Any:
        """
        Transforms bracket lookup into dotted access on the Namespace's
        mapping.
        """
        try:
            return self.__dict__[variable]
        except KeyError:
            return getattr(self, variable)

    def __delitem__(self, variable=None) -> None:
        """
        Deletes the item ``variable`` from the instance dictionary.
        """
        del self.__dict__[variable]

    async def __aiter__(self) -> t.AsyncGenerator[None, t.Hashable]:
        """
        Allows Namespace's to be unpacked with async iteration.
        """
        for variable in self.__dict__:
            await asyncio.sleep(0)
            yield variable

    def __iter__(self) -> t.Generator[None, t.Hashable, None]:
        """
        Allows Namespace's to be unpacked with tools like ``dict`` &
        ``list``.
        """
        yield from self.__dict__

    def keys(self) -> t.Generator[None, t.Hashable, None]:
        """
        Yields the names of all items in the instance.
        """
        yield from self.__dict__

    def values(self) -> t.Generator[None, t.Hashable, None]:
        """
        Yields the values of all items in the instance.
        """
        yield from self.__dict__.values()

    def items(self) -> t.Generator[None, t.Tuple[t.Hashable, t.Any], None]:
        """
        Yields the name, value pairs of all items in the instance.
        """
        yield from self.__dict__.items()


class OpenNamespace(Namespace):
    """
    A version of the `Namespace` class which doesn't omit instance
    repr's by default.
    """

    def __repr__(self, mask: bool = False) -> str:
        """
        Pretty displays the instance & its attributes.
        """
        return super().__repr__(mask=mask)


async def aimport_namespace(
    dictionary: dict, *, mapping: dict, deepcopy: bool = False
) -> None:
    """
    Takes a ``dictionary``, such as ``globals()``, and copies the
    key-value pairs from the ``mapping`` kwarg into it.
    """
    await asyncio.sleep(0)
    if deepcopy == True:
        dictionary.update(copy.deepcopy(mapping))
    else:
        dictionary.update(mapping)


def import_namespace(
    dictionary: dict, *, mapping: dict, deepcopy: bool = False
) -> None:
    """
    Takes a ``dictionary``, such as ``globals()``, and copies the
    key-value pairs from the ``mapping`` kwarg into it.
    """
    if deepcopy == True:
        dictionary.update(copy.deepcopy(mapping))
    else:
        dictionary.update(mapping)


async def amake_module(
    name: str, *, mapping: dict, deepcopy: bool = False
) -> Namespace:
    """
    Turns a mapping into a module object version of a Namespace which is
    importable using normal python syntax.
    """
    module = types.ModuleType(name)
    await aimport_namespace(
        module.__dict__, mapping=mapping, deepcopy=deepcopy
    )
    sys.modules[name] = module
    return Namespace(module.__dict__)


def make_module(
    name: str, *, mapping: dict, deepcopy: bool = False
) -> Namespace:
    """
    Turns a mapping into a module object version of a Namespace which is
    importable using normal python syntax.
    """
    module = types.ModuleType(name)
    import_namespace(module.__dict__, mapping=mapping, deepcopy=deepcopy)
    sys.modules[name] = module
    return Namespace(module.__dict__)


constants = make_module(
    "constants",
    mapping=dict(
        misc=make_module("misc", mapping=misc.__dict__),
        datasets=make_module("datasets", mapping=datasets.__dict__),
        passcrypt=make_module("passcrypt", mapping=passcrypt.__dict__),
        chunky2048=make_module("chunky2048", mapping=chunky2048.__dict__),
    ),
)


extras = dict(
    Namespace=Namespace,
    OpenNamespace=OpenNamespace,
    Slots=Slots,
    __doc__=__doc__,
    __package__=__package__,
    aimport_namespace=aimport_namespace,
    amake_module=amake_module,
    constants=constants,
    import_namespace=import_namespace,
    make_module=make_module,
)


commons = make_module("commons", mapping=extras, deepcopy=True)




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["_containers", "_exceptions", "_typing"]


__doc__ = (
    "Coordinates some of the library's UI/UX by inserting higher-level "
    "functionality from dependant modules into lower-level modules in t"
    "his package."
)


from . import *
from . import _typing, _containers, _exceptions
from ._exceptions import is_exception
from ._typing import Typing
from ._debuggers import gen_timer, agen_timer
from .asynchs import Processes, Threads, AsyncInit
from .commons import Namespace, OpenNamespace, Slots
from .commons import make_module
from .gentools import Comprende, comprehension
from .generics import Hasher, Domains, BytesIO, Padding, Clock
from .randoms import SequenceID
from .ciphers import StreamHMAC, SyntheticIV, Chunky2048
from .ciphers import generate_salt, agenerate_salt
from .keygens import KeyAADBundle
from .keygens import Ed25519, X25519, DomainKDF, Passcrypt
from .keygens import PackageSigner, PackageVerifier
from .databases import Database, AsyncDatabase


@comprehension()
async def adebugger(
    self: Comprende, *args, **kwargs
) -> Typing.AsyncGenerator:
    """
    Allows users to benchmark & read inspection details of running async
    generators inline as a chainable method.
    """
    args = args if args else self._args
    kwargs = {**self._kwargs, **kwargs}
    async for result in agen_timer(self._func)(*args, **kwargs):
        yield result


@comprehension()
def debugger(self: Comprende, *args, **kwargs) -> Typing.Generator:
    """
    Allows users to benchmark & read inspection details of running sync
    generators inline as a chainable method.
    """
    args = args if args else self._args
    kwargs = {**self._kwargs, **kwargs}
    yield from gen_timer(self._func)(*args, **kwargs)


def insert_debuggers() -> None:
    """
    Copies the addons over into the ``Comprende`` class.
    """
    addons = (debugger, adebugger)
    for addon in addons:
        setattr(Comprende, addon.__name__, addon)
        Comprende.lazy_generators.add(addon.__name__)


def insert_types() -> None:
    """
    Gives the package's type-hinting helper class access to the higher
    level classes.
    """
    Typing.add_type("AsyncCipherStream", AsyncCipherStream)
    Typing.add_type("AsyncDecipherStream", AsyncDecipherStream)
    Typing.add_type("AsyncInit", AsyncInit)
    Typing.add_type("CipherStream", CipherStream)
    Typing.add_type("DecipherStream", DecipherStream)
    Typing.add_type("AsyncDatabase", AsyncDatabase)
    Typing.add_type("BytesIO", BytesIO)
    Typing.add_type("Chunky2048", Chunky2048)
    Typing.add_type("Clock", Clock)
    Typing.add_type("Comprende", Comprende)
    Typing.add_type("Database", Database)
    Typing.add_type("DomainKDF", DomainKDF)
    Typing.add_type("Domains", Domains)
    Typing.add_type("Ed25519", Ed25519)
    Typing.add_type("GUID", GUID)
    Typing.add_type("Hasher", Hasher)
    Typing.add_type("KeyAADBundle", KeyAADBundle)
    Typing.add_type("Slots", Slots)
    Typing.add_type("Namespace", Namespace)
    Typing.add_type("OpenNamespace", OpenNamespace)
    Typing.add_type("PackageSigner", PackageSigner)
    Typing.add_type("PackageVerifier", PackageVerifier)
    Typing.add_type("Padding", Padding)
    Typing.add_type("Passcrypt", Passcrypt)
    Typing.add_type("PasscryptInstance", Passcrypt().__class__)
    Typing.add_type(
        "PasscryptSession", Passcrypt._passcrypt.__annotations__["session"]
    )
    Typing.add_type("Processes", Processes)
    Typing.add_type("SequenceID", SequenceID)
    Typing.add_type("StreamHMAC", StreamHMAC)
    Typing.add_type("SyntheticIV", SyntheticIV)
    Typing.add_type("Threads", Threads)
    Typing.add_type("X25519", X25519)
    for name, cls in _containers.__dict__.items():
        if (
            hasattr(cls, "mro")
            and issubclass(cls, Slots)
            and not hasattr(Typing, name)
        ):
            Typing.add_type(name, cls)
    for name, exc in _exceptions.__dict__.items():
        if is_exception(exc):
            Typing.add_type(name, exc)


def overwrite_containers_module() -> None:
    """
    Overwrites the package's `_containers.py` variable which will then
    be accessible to the user.
    """
    global _containers

    mapping = OpenNamespace(
        **{
            name: getattr(_containers, name) for name in _containers.__all__
        },
        __all__=_containers.__all__,
        __doc__=_containers.__doc__,
        __package__=_containers.__package__,
    )
    _containers = make_module("_containers", mapping=mapping)


def overwrite_exceptions_module() -> None:
    """
    Overwrites the package's `_exceptions.py` variable which will then
    be accessible to the user.
    """
    global _exceptions

    mapping = OpenNamespace(
        **{
            name: getattr(_exceptions, name) for name in _exceptions.__all__
        },
        __all__=_exceptions.__all__,
        __doc__=_exceptions.__doc__,
        __package__=_exceptions.__package__,
    )
    _exceptions = make_module("_exceptions", mapping=mapping)


def overwrite_typing_module() -> None:
    """
    Overwrites the package's `_typing.py` variable which will then be
    accessible to the user.
    """
    global _typing

    mapping = OpenNamespace(
        Typing=_typing.Typing,
        __all__=_typing.__all__,
        __doc__=_typing.__doc__,
        __package__=_typing.__package__,
    )
    _typing = make_module("_typing", mapping=mapping)


insert_debuggers()
insert_types()
overwrite_containers_module()
overwrite_exceptions_module()
overwrite_typing_module()




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__all__ = ["GUID", "acsprng", "csprng"]


__doc__ = (
    "A collection of functions which use & create varying levels of ent"
    "ropy for end-user cryptographic applications."
)


import math
import random as _random
from collections import deque
from secrets import choice, token_bytes
from secrets import randbits as token_bits
from hashlib import sha3_256, sha3_512, shake_256
from .__constants import *
from ._exceptions import *
from ._containers import UnmaskedGUID
from ._typing import Typing as t
from .commons import OpenNamespace
from .commons import make_module
from .asynchs import Threads
from .asynchs import (
    asleep,
    asyncio,
    gather,
    sleep,
    run,
    ns_time,
    ns_counter,
)
from .paths import SecurePath
from .paths import read_salt_file
from .generics import Domains, Hasher, Clock, BytesIO
from .generics import bytes_as_int, int_as_bytes


async def auniform(*a, **kw) -> float:
    """
    Asynchronous version of the standard library's `random.uniform`.
    """
    await asleep()
    return uniform(*a, **kw)


async def achoice(iterable: t.Iterable[t.Any]) -> t.Any:
    """
    Asynchronous version of the standard library's `secrets.choice`.
    """
    await asleep()
    return choice(iterable)


async def aunique_range(*a, **kw) -> int:
    """
    Asynchronous version of the standard library's `random.randrange`.
    """
    await asleep()
    return unique_range(*a, **kw)


async def arandom_sleep(span: t.PositiveRealNumber = 2) -> None:
    """
    Asynchronously sleeps for a psuedo-random portion of ``span`` time,
    measured in seconds.
    """
    await asleep(span * uniform(0, 1))


def random_sleep(span: t.PositiveRealNumber = 2) -> None:
    """
    Synchronously sleeps for a psuedo-random portion of ``span`` time,
    measured in seconds.
    """
    sleep(span * uniform(0, 1))


async def atoken_bits(size: int) -> int:
    """
    Returns ``size`` number of bits from `secrets.randbits`.
    """
    await asleep()
    return token_bits(size)


async def atoken_bytes(size: int) -> bytes:
    """
    Returns ``size`` bytes of `secrets.token_bytes` entropy.
    """
    await asleep()
    return token_bytes(size)


async def _asalt_multiply(*numbers: t.Iterable[int]) -> int:
    """
    Allows for non-commutative multiplication. This assists pseudo-
    random number generators in turning combinations of low entropy
    number sources into permutations. This greatly increases the amount
    of knowledge an attacker must have to perform pre-image or known-
    plaintext attacks on the calculations usings the hashes of those
    permutations.
    """
    global _mix, _mod, _seed, _offset

    _mix ^= abs(sum((_seed, _offset, ns_time(), ns_counter(), *numbers)))
    mix = _mix = _mix % _mod
    _seed ^= mix
    start = _seed
    _offset ^= _seed
    await asleep()
    for number in numbers:
        mix += _offset
        start *= number ^ mix
    await asleep()
    return start ^ _seed


def _salt_multiply(*numbers: t.Iterable[int]) -> int:
    """
    Allows for non-commutative multiplication. This assists pseudo-
    random number generators in turning combinations of low entropy
    number sources into permutations. This greatly increases the amount
    of knowledge an attacker must have to perform pre-image or known-
    plaintext attacks on the calculations usings the hashes of those
    permutations.
    """
    global _mix, _mod, _seed, _offset

    _mix ^= abs(sum((_seed, _offset, ns_time(), ns_counter(), *numbers)))
    mix = _mix = _mix % _mod
    _seed ^= mix
    start = _seed
    _offset ^= _seed
    for number in numbers:
        mix += _offset
        start *= number ^ mix
    return start ^ _seed


class EntropyDaemon:
    """
    Creates & manages a background thread which asynchronously extracts
    from & seeds new entropy into this module's entropy pools. Mixing
    background threading, asynchrony, pseudo random sleeps & sha3_512
    hashing with a single, shared global object results in highly
    unpredictable & non-deterministic effects on entropy generation for
    the whole package.
    """

    __slots__ = (
        "_cancel",
        "_currently_mutating_frequency",
        "_daemon",
        "_frequency",
        "_initial_frequency",
        "_pool",
    )

    def __init__(
        self,
        entropy_pool: t.Deque[bytes],
        *,
        frequency: t.PositiveRealNumber = 1,
    ) -> "self":
        """
        Prepares an instance to safely start a background thread.
        """
        self._pool = entropy_pool
        self._daemon = None
        self._cancel = False
        self._currently_mutating_frequency = False
        self.set_frequency(frequency)

    async def _anew_snapshot(self) -> bytes:
        """
        Returns 144-bytes of pseudo-random values from the instance's
        entropy pool & the `secrets.token_bytes` function.
        """
        await asleep()
        return await atoken_bytes(32) + self._pool[0] + self._pool[-1][:48]

    def _set_temporary_frequency(
        self,
        frequency: t.PositiveRealNumber,
        duration: t.PositiveRealNumber,
    ) -> None:
        """
        Sets a temporary maximum number of seconds a started entropy
        daemon will pseudo-randomly sleep in-between each iteration. The
        frequency will return to its initial value after ``duration``
        number of seconds.
        """
        try:
            self._frequency = frequency
            sleep(duration)
        finally:
            self._frequency = self._initial_frequency
            self._currently_mutating_frequency = False

    def set_temporary_frequency(
        self,
        frequency: t.PositiveRealNumber = 0.001,
        *,
        duration: t.PositiveRealNumber = 1,
    ) -> "self":
        """
        Sets a temporary maximum number of seconds a started entropy
        daemon will pseudo-randomly sleep in-between each iteration. The
        frequency will return to its initial value after ``duration``
        number of seconds.
        """
        if not self._currently_mutating_frequency:
            self._currently_mutating_frequency = True
            Threads.submit(
                self._set_temporary_frequency, frequency, duration
            )
        return self

    def set_frequency(
        self, frequency: t.PositiveRealNumber = 1
    ) -> "self":
        """
        Sets the maximum number of seconds a started entropy daemon will
        pseudo-randomly sleep in-between each iteration. Setting the
        ``frequency`` to smaller numbers will cause more cpu power &
        to be consumed by the background daemon thread.
        """
        self._frequency = frequency
        self._initial_frequency = frequency
        return self

    async def _araw_loop(self) -> None:
        """
        Takes snapshots of & feeds entropy into the module's entropy
        pools before & after sleeping for a pseudo-random amount of time.
        This is done asynchronously & in a background thread to increase
        the unpredictability & non-determinism of entropy generation.
        """
        while True:
            seed = await self._anew_snapshot()
            await _add_to_pool(await _entropy.ahash(seed), self._pool)
            _xof.update(self._pool[0])
            await arandom_sleep(self._frequency)
            await _add_to_pool(await _entropy.ahash(seed), self._pool)
            _xof.update(self._pool[0])
            if self._cancel:
                break

    def start(self) -> "self":
        """
        Runs an entropy updating & gathering thread in the background.

        This supports the package by asynchronously & continuously
        seeding into & extracting new entropy from its entropy pools.
        """
        state = Threads._state_machine().list()
        self._daemon = Threads._type(
            target=Threads._run_async_func,
            args=[self._araw_loop],
            kwargs=dict(_state=state),
        )
        self._daemon.daemon = True
        self._daemon.start()
        return self

    def cancel(self) -> "self":
        """
        Cancels the background thread.
        """
        self._cancel = True
        return self


#  initialize a global entropy pool
_pool = deque([token_bytes(64), token_bytes(64)], maxlen=32)


async def _add_to_pool(
    entropy: bytes, entropy_pool: t.Sequence[bytes] = _pool
) -> None:
    """
    Prevents writes of the same `_entropy.hash` outputs to the global
    entropy pool.
    """
    while entropy in entropy_pool:
        entropy = _entropy.hash(entropy)
        await arandom_sleep(0.001)
    entropy_pool.appendleft(entropy)


# initialize the global hashing objects that also collect entropy
_entropy = Hasher(token_bytes(304) + b"".join(_pool))
_xof = Hasher(token_bytes(208) + b"".join(_pool), obj=shake_256)

# avert event loop clashes
run = asyncio.new_event_loop().run_until_complete

# initializing weakly entropic functions
random = _random.Random(token_bytes(2500))
uniform = random.uniform
unique_range = random.randrange

_mod = PRIMES[256][-1]
_offset = token_bits(256)
_mix = int.from_bytes(_entropy.hash(*_pool), BIG)
_seed = int.from_bytes(_entropy.hash(*_pool)[:32], BIG)
_numbers = (_mix, _seed, _offset)

_ = _salt_multiply(*_numbers)
run(_asalt_multiply(_, *_numbers))

_initial_entropy = deque([token_bits(1024), token_bits(1024)], maxlen=2)

# ensure the device has created a static salt for GUID creation
_static_salt_name = Domains.encode_constant(b"default_guid_name", size=16)
_device_salt_path = SecurePath(key=_static_salt_name, _admin=True)
_device_salt = read_salt_file(_device_salt_path)

# begin the entropy gathering daemon
_entropy_daemon = EntropyDaemon(_pool).start()
_entropy_daemon.set_temporary_frequency(0.001, duration=2)


async def _asalt() -> int:
    """
    Returns a low-grade entropy number from cached & ratcheted system
    entropy.
    """
    entropy = _entropy.hash(token_bytes(32), *_pool)[:32]
    _initial_entropy.appendleft(int.from_bytes(entropy, BIG))
    return await _asalt_multiply(*_initial_entropy)


def _salt() -> int:
    """
    Returns a low-grade entropy number from cached & ratcheted system
    entropy.
    """
    entropy = _entropy.hash(token_bytes(32), *_pool)[:32]
    _initial_entropy.appendleft(int.from_bytes(entropy, BIG))
    return _salt_multiply(*_initial_entropy)


async def arandom_number_generator(
    size: int = 64,
    *,
    entropy: t.Any = _entropy.hash(run(_asalt()).to_bytes(2048, BIG)),
    freshness: int = 8,
) -> bytes:
    """
    Returns a ``size``-byte random bytestring derived from very large
    entropy pools, the provided ``entropy`` value, persistent sha3
    hashing objects, & a multithreaded + asynchronous chaotic algorithm.
    The number of ``freshness`` dictates how long the chain of
    asynchronous tasks will be set in motion gathering entropy to seed
    the output function.

     _____________________________________
    |                                     |
    |        Algorithm Explanation:       |
    |_____________________________________|

    We propose several methods for producing cryptographically secure
    pseudo-random numbers, and implement them in this function. This
    function is implemented to securely produce entropy under threat
    models that don't include attackers that can maliciously analyze
    users' machines' memory. Such a capability represents a serious
    threat to the unpredictability of the CSPRNG, as it suggests the
    ability to read any arbitrary secrets stored by the machine.

    0. Use many component PRNGs with differing designs & sources to
    spread the risk of any particularly weak ones & benefit from the
    most reliable ones by securely mixing the entropy bits they may
    produce together.

    1. Make the search space for an attacker to guess the inputs to any
    of the component PRNGs intractibly large by using seeds at least as
    large as the minumum acceptable security strength.

    2. Make new seeds after &/or before each use, import-time &/or
    function call.

    3. Use cryptographic hashing on the inputs & outputs of component
    PRNGs to unlink them from their internal states.

    4. Incorporate forward-secure key ratcheting algorithms at key
    points along the PRNGs' communications routes with themselves & the
    user.

    5. Use the powerful sha3_256, sha3_512 or shake_256 algorithms for
    all hashing.

    6. Persist & use a sha3_512 & shake_256 hashing object for hidden
    &/or internal procedures across the module for the automatic & non-
    deterministic collection of entropy throughout normal use of the
    module.

    7. Further frustrate an attacker by necessitating that they perform
    an order prediction attack on the results of the CSPRNG's component
    PRNGs. We do this by using asynchrony entirely throughout the CSPRNG
    with random sleeps, and using a non-commutative, salted, modular
    multiplication to combine internal states into a single element in
    its ratcheting internal state.

    8. Allow the user to securely mix any extra entropy they have access
    to into the CSPRNG's state machine. This mitigates some impacts of
    the malicious analysis of memory by an adversary.

    9. Use a background thread which continuously hashes & updates two
    of the package's entropy pools with new entropic material & their
    internal states. This adds unpredictable alterations to the pools
    concurrently with the running of the package.

    10. Iterate and interleave all these methods enough times such that,
    if we assume each time the CSPRNG produces a 64-byte internal state,
    that only 8-bits of entropy are produced. Then, initializing up to
    a cache of 32 ratcheting states, the ``random_number_generator``
    algorithm here would have at least 256-bits of entropy.

    **** **** **** **** **** **** **** **** **** **** **** **** ****

    Our implementation analysis is NOT rigorous or conclusive. Though
    soley based on the constraints stated above, our assumptions of
    randomness should be reasonable estimates within the threat model
    where an attacker cannot arbitrarily analyze users' machines'
    memory.

    **** **** **** **** **** **** **** **** **** **** **** **** ****
    """
    xof = _xof.copy()
    _entropy_daemon.set_temporary_frequency(0.001, duration=1)

    if entropy.__class__ is not bytes:
        entropy = repr(entropy).encode()

    if freshness or not _pool:

        async def create_unique_multiple(seed: int) -> int:
            return await _asalt_multiply(size, seed, token_bits(256))

        async def big_modulation(*args) -> int:
            return await _asalt_multiply(
                size, *args, await atoken_bits(256)
            ) % await achoice(PRIMES[512])

        async def modular_multiplication() -> None:
            seed = await _asalt() % await achoice(PRIMES[512])
            await arandom_sleep(0.003)
            multiples = (create_unique_multiple(seed) for _ in range(3))
            multiples = [await multiple for multiple in multiples]
            result = await big_modulation(seed, *multiples)
            await _add_to_pool(
                await _entropy.ahash(
                    seed.to_bytes(64, "big"), result.to_bytes(64, "big")
                )
            )

        async def add_to_pool() -> None:
            seed = await xof.ahash(await atoken_bytes(32), size=32)
            await arandom_sleep(0.003)
            await _add_to_pool(await _entropy.ahash(entropy, seed))

        async def start_generator() -> None:
            tasks = deque()
            rounds = 8 if not freshness else freshness
            for _ in range(rounds):
                await asleep()
                tasks.appendleft(modular_multiplication())
                for _ in range(10):
                    tasks.appendleft(add_to_pool())
            await gather(
                *sorted(tasks, key=lambda val: token_bytes(32)),
                return_exceptions=True,
            )

        await start_generator()
    else:
        await _add_to_pool(
            await _entropy.ahash(await atoken_bytes(32), entropy)
        )

    # Prevent the possibility that multiple threads will retrieve the
    # same result if they each happen to interrupt each other multiple
    # times in between their update of _xof & their call for its digest
    # by using a unique copy.
    return await xof.ahash(token_bytes(xof.block_size), *_pool, size=size)


def random_number_generator(
    size: int = 64,
    *,
    entropy: t.Any = _entropy.hash(_salt().to_bytes(2048, BIG)),
    freshness: int = 8,
) -> bytes:
    """
    Returns a ``size``-byte random bytestring derived from very large
    entropy pools, the provided ``entropy`` value, persistent sha3
    hashing objects, & a multithreaded + asynchronous chaotic algorithm.
    The number of ``freshness`` dictates how long the chain of
    asynchronous tasks will be set in motion gathering entropy to seed
    the output function.

     _____________________________________
    |                                     |
    |        Algorithm Explanation:       |
    |_____________________________________|

    We propose several methods for producing cryptographically secure
    pseudo-random numbers, and implement them in this function. This
    function is implemented to securely produce entropy under threat
    models that don't include attackers that can maliciously analyze
    users' machines' memory. Such a capability represents a serious
    threat to the unpredictability of the CSPRNG, as it suggests the
    ability to read any arbitrary secrets stored by the machine.

    0. Use many component PRNGs with differing designs & sources to
    spread the risk of any particularly weak ones & benefit from the
    most reliable ones by securely mixing the entropy bits they may
    produce together.

    1. Make the search space for an attacker to guess the inputs to any
    of the component PRNGs intractibly large by using seeds at least as
    large as the minumum acceptable security strength.

    2. Make new seeds after &/or before each use, import-time &/or
    function call.

    3. Use cryptographic hashing on the inputs & outputs of component
    PRNGs to unlink them from their internal states.

    4. Incorporate forward-secure key ratcheting algorithms at key
    points along the PRNGs' communications routes with themselves & the
    user.

    5. Use the powerful sha3_256, sha3_512 or shake_256 algorithms for
    all hashing.

    6. Persist & use a sha3_512 & shake_256 hashing object for hidden
    &/or internal procedures across the module for the automatic & non-
    deterministic collection of entropy throughout normal use of the
    module.

    7. Further frustrate an attacker by necessitating that they perform
    an order prediction attack on the results of the CSPRNG's component
    PRNGs. We do this by using asynchrony entirely throughout the CSPRNG
    with random sleeps, and using a non-commutative, salted, modular
    multiplication to combine internal states into a single element in
    its ratcheting internal state.

    8. Allow the user to securely mix any extra entropy they have access
    to into the CSPRNG's state machine. This mitigates some impacts of
    the malicious analysis of memory by an adversary.

    9. Use a background thread which continuously hashes & updates two
    of the package's entropy pools with new entropic material & their
    internal states. This adds unpredictable alterations to the pools
    concurrently with the running of the package.

    10. Iterate and interleave all these methods enough times such that,
    if we assume each time the CSPRNG produces a 64-byte internal state,
    that only 8-bits of entropy are produced. Then, initializing up to
    a cache of 32 ratcheting states, the ``random_number_generator``
    algorithm here would have at least 256-bits of entropy.

    **** **** **** **** **** **** **** **** **** **** **** **** ****

    Our implementation analysis is NOT rigorous or conclusive, though
    soley based on the constraints stated above, our assumptions of
    randomness should be reasonable estimates within the threat model
    where an attacker cannot arbitrarily analyze users' machines'
    memory.

    **** **** **** **** **** **** **** **** **** **** **** **** ****
    """
    xof = _xof.copy()
    _entropy_daemon.set_temporary_frequency(0.001, duration=1)

    if entropy.__class__ is not bytes:
        entropy = repr(entropy).encode()

    if freshness or not _pool:

        async def create_unique_multiple(seed: int) -> int:
            return await _asalt_multiply(size, seed, token_bits(256))

        async def big_modulation(*args) -> int:
            return await _asalt_multiply(
                size, *args, await atoken_bits(256)
            ) % await achoice(PRIMES[512])

        async def modular_multiplication() -> None:
            seed = await _asalt() % await achoice(PRIMES[512])
            await arandom_sleep(0.003)
            multiples = (create_unique_multiple(seed) for _ in range(3))
            multiples = [await multiple for multiple in multiples]
            result = await big_modulation(seed, *multiples)
            await _add_to_pool(
                await _entropy.ahash(
                    seed.to_bytes(64, "big"), result.to_bytes(64, "big")
                )
            )

        async def add_to_pool() -> None:
            seed = await xof.ahash(await atoken_bytes(32), size=32)
            await arandom_sleep(0.003)
            await _add_to_pool(await _entropy.ahash(entropy, seed))

        async def start_generator() -> None:
            tasks = deque()
            rounds = 8 if not freshness else freshness
            for _ in range(rounds):
                await asleep()
                tasks.appendleft(modular_multiplication())
                for _ in range(10):
                    tasks.appendleft(add_to_pool())
            await gather(
                *sorted(tasks, key=lambda val: token_bytes(32)),
                return_exceptions=True,
            )

        run(start_generator())  # <- RuntimeError in event loops
    else:
        run(_add_to_pool(_entropy.hash(token_bytes(32), entropy)))

    # Prevent the possibility that multiple threads will retrieve the
    # same result if they each happen to interrupt each other multiple
    # times in between their update of _xof & their call for its digest
    # by using a unique copy.
    return xof.hash(token_bytes(xof.block_size), *_pool, size=size)


async def agenerate_salt(size: int = SALT_BYTES) -> bytes:
    """
    Returns ``size`` random bytes for use as an ephemeral, uniform salt.
    """
    if size < 8 or size > 64:
        raise Issue.invalid_value("salt size", "<8 or >64")
    return await atoken_bytes(size)


def generate_salt(size: int = SALT_BYTES) -> bytes:
    """
    Returns ``size`` random bytes for use as an ephemeral, uniform salt.
    """
    if size < 8 or size > 64:
        raise Issue.invalid_value("salt size", "<8 or >64")
    return token_bytes(size)


async def agenerate_iv(size: int = IV_BYTES) -> bytes:
    """
    Returns ``size`` random bytes for use as an ephemeral, uniform IV.
    """
    if size < 8 or size > 64:
        raise Issue.invalid_value("iv size", "<8 or >64")
    return await atoken_bytes(size)


def generate_iv(size: int = IV_BYTES) -> bytes:
    """
    Returns ``size`` random bytes for use as an ephemeral, uniform IV.
    """
    if size < 8 or size > 64:
        raise Issue.invalid_value("iv size", "<8 or >64")
    return token_bytes(size)


async def agenerate_siv_key(size: int = SIV_KEY_BYTES) -> bytes:
    """
    Returns ``size`` random bytes for use as an ephemeral, uniform
    SIV-key.
    """
    if size < 8 or size > 64:
        raise Issue.invalid_value("siv_key size", "<8 or >64")
    return await atoken_bytes(size)


def generate_siv_key(size: int = SIV_KEY_BYTES) -> bytes:
    """
    Returns ``size`` random bytes for use as an ephemeral, uniform
    SIV-key.
    """
    if size < 8 or size > 64:
        raise Issue.invalid_value("siv_key size", "<8 or >64")
    return token_bytes(size)


async def agenerate_raw_guid(
    size: int = GUID_BYTES,
    *,
    clock: Clock = Clock(NANOSECONDS, epoch=EPOCH_NS),
    timestamp_bytes: int = 8,
) -> bytes:
    """
    Returns a raw ``size``-byte globally unique identifier. The first
    8-bytes are a nanosecond timestamp, & the rest are random bytes. A
    custom ``clock`` can be specified, which produces timestamps from
    its `amake_timestamp` method.
    --------
    WARNING: DO NOT use if timestamp information, for security or
    -------- anonymity purposes, is supposed to be kept secret from the
    contexts where these guids are exposed. One example: creating a
    guid, then performing an operation on some sensitive or secret
    information, & then exposing the guid, could leak information about
    those secrets to an observer calculating the time difference between
    the timestamp & the moment it was exposed. DO NOT DO THIS. USE WITH
    CAUTION, in contexts where the guids remain secret or where secret
    information is NOT being touched AND the time information DOES NOT
    harm anonymity or security!
    """
    if size < 10 or size > 64:
        raise Issue.invalid_value("guid size", "<10 or >64")
    return (
        await clock.amake_timestamp(size=timestamp_bytes)
        + await atoken_bytes(size - timestamp_bytes)
    )


def generate_raw_guid(
    size: int = GUID_BYTES,
    *,
    clock: Clock = Clock(NANOSECONDS, epoch=EPOCH_NS),
    timestamp_bytes: int = 8,
) -> bytes:
    """
    Returns a raw ``size``-byte globally unique identifier. The first
    8-bytes are a nanosecond timestamp, & the rest are random bytes. A
    custom ``clock`` can be specified, which produces timestamps from
    its `make_timestamp` method.
    --------
    WARNING: DO NOT use if timestamp information, for security or
    -------- anonymity purposes, is supposed to be kept secret from the
    contexts where these guids are exposed. One example: creating a
    guid, then performing an operation on some sensitive or secret
    information, & then exposing the guid, could leak information about
    those secrets to an observer calculating the time difference between
    the timestamp & the moment it was exposed. DO NOT DO THIS. USE WITH
    CAUTION, in contexts where the guids remain secret or where secret
    information is NOT being touched AND the time information DOES NOT
    harm anonymity or security!
    """
    if size < 10 or size > 64:
        raise Issue.invalid_value("guid size", "<10 or >64")
    return (
        clock.make_timestamp(size=timestamp_bytes)
        + token_bytes(size - timestamp_bytes)
    )


class SequenceID:
    """
    A class for producing unique, deterministic pseudo-random sequential
    identifiers that do not suffer from birthday bound collisions if all
    input indexes fall within two multiples of the prime used for the
    size of outputs specified by the user. Each prime used for each
    output-size is the largest possible prime of that byte-size. Each
    primitive root used for each prime is the smallest prime primitive
    root >= 7 for that prime.

    The produced identifiers are randomized by unique, secret, uniform
    salts. Any instance created using the same salt is able to create
    the same sequence of identifiers. Normal birthday-bound collision
    probabilities apply when different salts are used between callers.

     _____________________________________
    |                                     |
    |            Usage Example:           |
    |_____________________________________|

    from aiootp.randoms import SequenceID, token_bytes
    from aiootp.ciphers import ChaCha20Poly1305
    from aiootp.keygens import DomainKDF


    seed = token_bytes(32)
    kdf = DomainKDF(b"movie-collection", seed, key=session.shared_key)
    cipher = ChaCha20Poly1305(kdf.sha3_256(context=b"server-encryption-key"))
    sid = SequenceID(
        salt=kdf.shake_256(12, context=b"server-sequence-ids"), size=12
    )

    yield seed
    for i, movie in enumerate(movie_collection):
        yield cipher.encrypt(sid.new(i), movie, session.associated_data)

    """

    __slots__ = ("_size", "_gen", "_prime", "_subprime", "_salt", "_key")

    _IPAD: bytes = b"\xa3"
    _OPAD: bytes = b"\x1b"
    _XPAD: bytes = b"\x8d"
    _MIN_SIZE: int = 4
    _MAX_SIZE: int = 32
    _MIN_SALT_SIZE: int = 16

    _primes: PrimeGroups = PrimeGroups()

    @classmethod
    def _encode_salt(
        cls, salt: bytes, prime: int, size: int
    ) -> t.Tuple[int, int]:
        """
        Returns three integer copies of a bytes-type ``salt`` value
        xored with different constants which have an optimal hamming
        distance of half their bit-length. The first two copies returned
        are not multiples of the ``prime`` & are one byte longer than
        the original salt. The last copy is ``size`` bytes long.
        """
        subprime = cls._primes[f"o{8 * size}"]
        pad_size = len(salt) + 1
        integer_salt = bytes_as_int(salt)
        short_integer_salt = bytes_as_int(salt[:size])
        isalt = bytes_as_int(pad_size * cls._IPAD) ^ integer_salt
        osalt = bytes_as_int(pad_size * cls._OPAD) ^ integer_salt
        xsalt = bytes_as_int(size * cls._XPAD) ^ short_integer_salt
        while (not isalt % prime) or (not isalt % subprime):
            isalt += 1
        while (not osalt % prime) or (not osalt % subprime):
            osalt += 1
        return isalt, osalt, xsalt

    def __init__(self, salt: bytes, *, size: int = 12) -> "self":
        """
        Prepares the instance to run the algorithm after checking that
        the provided ``size`` of outputs & uniform ``salt`` are
        supported & work together to safely run the algorithm.
        """
        self._install_configuration(size)
        self._install_salt(salt)
        self._install_obfuscator()

    @property
    def _session_configuration(self) -> t.Tuple[int, int, int, int]:
        """
        A utility for improving the readability of getting access to all
        the configuartion values of the instance.
        """
        return self._size, self._gen, self._prime, self._subprime

    def _install_configuration(self, size: int) -> None:
        """
        Sets the prime & associated values which form the groups used
        for the specified ``size`` of outputs.
        """
        if size.__class__ is not int:
            raise Issue.value_must_be_type("size", int)
        elif size < self._MIN_SIZE or size > self._MAX_SIZE:
            raise Issue.value_must(
                "size", f"be >= {self._MIN_SIZE} and <= {self._MAX_SIZE}"
            )
        bit_size = 8 * size
        self._size = size
        self._gen = self._primes[f"g{bit_size}"]
        self._prime = self._primes[f"p{bit_size}"]
        self._subprime = self._primes[f"o{bit_size}"]

    def _install_salt(self, salt: bytes) -> None:
        """
        Stores a uniform bytes-type ``salt`` after checking if it's
        the correct type & is sufficiently large to safely work with the
        algorithm.
        """
        min_size = max((self._size, self._MIN_SALT_SIZE))
        if salt.__class__ is not bytes:
            raise Issue.value_must_be_type("salt", bytes)
        elif len(salt) < min_size:
            raise Issue.value_must("len(salt)", f"be >= {min_size}")
        self._salt = salt

    def _install_obfuscator(self) -> None:
        """
        Stores a group element generator function which ensures each
        output is unique for all inputs between two multiples of the
        isntance's subprime. The inputs & outputs are also obfuscated by
        the secrecy of the user-defined salt.
        """
        size, gen, prime, subprime = self._session_configuration
        isalt, osalt, xsalt = self._encode_salt(self._salt, prime, size)
        self._key = lambda index: (
            xsalt ^ pow(gen, (osalt * (isalt + index)) % subprime, prime)
        ).to_bytes(size, BIG)

    async def anew(
        self,
        index: int,
        *,
        encode: bool = False,
        encoder: t.Callable[..., bytes] = BytesIO.bytes_to_urlsafe,
    ) -> bytes:
        """
        Produces a raw-bytes pseudo-random identifier that is guaranteed
        to be unique if each call uses a different ``index``, each
        ``index`` is congruent to a distinct sequential natural number
        from a start index to an end index whose difference is smaller
        than the instance's subprime (determined by the instance `size`
        parameter), & each is produced using the same instance `salt`.
        The produced identifiers are randomized by unique, secret,
        uniform salts. Any instance created using the same salt is able
        to create the same sequence of identifiers. Normal birthday-
        bound collision probabilities apply when different salts are
        used between callers.

        If ``encode`` is set to `True`, then the result is first passed
        as an argument to the ``encoder`` callable before being returned.
        By default, the ``encoder`` transforms the raw bytes into a url
        safe base64 value without b"=" characters.
        """
        await asleep()
        if index < 0:
            raise Issue.value_must("index", "be >= 0")
        elif encode:
            return encoder(self._key(index))
        return self._key(index)

    def new(
        self,
        index: int,
        *,
        encode: bool = False,
        encoder: t.Callable[..., bytes] = BytesIO.bytes_to_urlsafe,
    ) -> bytes:
        """
        Produces a raw-bytes pseudo-random identifier that is guaranteed
        to be unique if each call uses a different ``index``, each
        ``index`` is congruent to a distinct sequential natural number
        from a start index to an end index whose difference is smaller
        than the instance's subprime (determined by the instance `size`
        parameter), & each is produced using the same instance `salt`.
        The produced identifiers are randomized by unique, secret,
        uniform salts. Any instance created using the same salt is able
        to create the same sequence of identifiers. Normal birthday-
        bound collision probabilities apply when different salts are
        used between callers.

        If ``encode`` is set to `True`, then the result is first passed
        as an argument to the ``encoder`` callable before being returned.
        By default, the ``encoder`` transforms the raw bytes into a url
        safe base64 value without b"=" characters.
        """
        if index < 0:
            raise Issue.value_must("index", "be >= 0")
        elif encode:
            return encoder(self._key(index))
        return self._key(index)


class GUID(SequenceID):
    r"""
    A class for producing pseudo-random identifiers that are guaranteed
    to be unique if all calls occur on a different nanosecond & use the
    same instance `salt`. Additionally, any calls which utilize the same
    `salt` but a different `node_number` will always produce unique
    outputs from each other, even if they occur on the same nanosecond.
    Also, outputs are guaranteed to be unique for any calls from any
    unique instance that utilizes the same `salt` & `node_number` if
    no more than 256 calls are made every nanosecond. The probability of
    per-nanosecond uniqueness for instances using the same `salt` can be
    increased exponentially with linear increases to the instance `size`
    parameter.

    Normal birthday-bound collision probabilities apply when different
    salts are used between callers.

    All outputs are also invertable. This means outputs can be unmasked
    & sorted according to `timestamp`, `node_number` & `counter`.

    By default only 256 different `node_number`s are supported, as they
    are represented in one byte, but as demonstrated in an example below,
    subclasses can customize how many bytes to allocate for node numbers.
    --------
    WARNING: The produced identifiers are randomized & obfuscated but
    -------- they are also invertible. The user must beware not to
    expose the guids if the `node_number`, `salt` or the current time in
    nanoseconds must remain secret from the audiences able to view the
    guids.

     _____________________________________
    |                                     |
    |     Usage Example: Primary Keys     |
    |_____________________________________|

    import aiootp

    dds = distributed_database_system
    shared_salt = dds.shared_salt
    guid = aiootp.GUID(shared_salt, node_number=local_db.id, size=16)

    for table in local_db.tables:
        for record in table.new_records:
            record.dds_primary_key = guid.new()
            dds.merge(table, record)

     _____________________________________
    |                                     |
    |     Usage Example: Node Numbers     |
    |_____________________________________|

    import aiootp

    class LargeNetworkGUID(aiootp.GUID, node_number_bytes=4):
        pass

    for i, node in enumerate(network.nodes):  # 256**4 possible nodes
        node.install_guid_generator(
            guid=LargeNetworkGUID(network.salt, node_number=i, size=16)
        )

     _____________________________________
    |                                     |
    |  Usage Example: Configuration Demo  |
    |_____________________________________|

    import aiootp
    from io import StringIO

    # 16 raw bytes without encoding ->
    guid = aiootp.GUID(size=16)
    print(repr(guid.new()))
    b'x]\xd2\xe6\x0c\xb2F\xf9\x05\x02\xa9\xf1\x84\xa3\x0c&'

    # urlsafe base64 encoding ->
    print(repr(guid.new(encode=True)))
    b'3umIg-hZ2J_g8malraZpnw'

    # custom encoding ->
    TRADITIONAL_LAYOUT = (8, 4, 4, 4, 12)

    def hex_segment_guid(guid: bytes) -> str:
        read = StringIO(guid.hex()).read
        return "-".join(read(size) for size in TRADITIONAL_LAYOUT)

    print(repr(guid.new(encode=True, encoder=hex_segment_guid)))
    '5e152c91-2f27-8ac6-fa4c-4041ba23d93d'
    """

    __slots__ = ("_node_number", "_offset_npad", "_unmask")

    _NODE_NUMBER_BYTES: int = NODE_NUMBER_BYTES
    _NPAD = int.from_bytes(NODE_NUMBER_BYTES * b"i", BIG)

    _COUNTER_BYTES: int = 1
    _MIN_SIZE: int = MIN_GUID_BYTES
    _MIN_RAW_SIZE: int = MIN_RAW_GUID_BYTES
    _SAFE_TIMESTAMP_BYTES: int = SAFE_TIMESTAMP_BYTES

    def __init_subclass__(
        cls, node_number_bytes: int = NODE_NUMBER_BYTES, **kw
    ) -> None:
        """
        Allows subclasses to only define a number of bytes to assign for
        node numbers so the padding it's xored with will be automatically
        extended to match.
        """
        super().__init_subclass__(**kw)
        cls._NODE_NUMBER_BYTES = node_number_bytes
        cls._NPAD = int.from_bytes(node_number_bytes * b"i", BIG)

    def __init__(
        self,
        salt: t.Optional[bytes] = None,
        *,
        size: int = GUID_BYTES,
        node_number: int = 0,
    ) -> "self":
        """
        Prepares the instance to run the algorithm after checking that
        the provided ``size`` of outputs & ``salt`` are supported & work
        together to safely run the algorithm. If no ``salt`` is provided,
        then the automatically generated salt which is stored on the
        device is used.
        """
        salt = salt if salt else _device_salt
        self._install_node_number(node_number, size=size)
        super().__init__(salt=salt, size=size)

    def _install_node_number(self, node_number: int, size: int) -> None:
        """
        Stores an 'N'-byte ``node_number`` that's xor'd with a constant
        which shares an optimal hamming distance of half its bit-length
        with the class' `_IPAD` & `_OPAD` values. 'N' is the class'
        `_NODE_NUMBER_BYTES` value, which assigns a size (& maximum
        possible unique values) for the node_number. By default 'N' is
        1, & subclasses can customize this parameter in the subclass
        initializer.
        """
        node_bytes = self._NODE_NUMBER_BYTES
        min_size = node_bytes + self._MIN_RAW_SIZE + self._COUNTER_BYTES
        if size < min_size:
            raise Issue.value_must(
                "size", f"be at least {min_size} to fit the node number"
            )
        self._offset_npad = self._NPAD << 8
        self._node_number = int_as_bytes(
            node_number ^ self._NPAD, size=node_bytes
        )

    def _obfuscator_shortcuts(
        self, size: int, prime: int, isalt: int
    ) -> t.Tuple[int, int, int, bytes, callable, int]:
        """
        Aggregates the calculated values used to run the algorithm for
        efficient referencing.
        """
        guid_size = size - self._NODE_NUMBER_BYTES - self._COUNTER_BYTES
        offset_npad = self._offset_npad
        node = self._node_number
        _int = int.from_bytes
        inverse = pow(isalt, prime - 2, prime)
        return guid_size, offset_npad, node, _int, inverse

    def _install_obfuscator(self) -> None:
        """
        Stores an efficient guid generator function that obfuscates the
        outputs of affine-group operations on the user-defined salt, a
        user-defined node number, a nanosecond-time & random-bytes raw
        guid, & a 1-byte counter.
        """
        def counter() -> bytes:
            nonlocal i

            i = (i + 1) % 256
            return i.to_bytes(1, BIG)

        i, size, _, prime, _ = 0, *self._session_configuration
        isalt, osalt, xsalt = self._encode_salt(self._salt, prime, size)
        (
            guid_size, offset_npad, node, _int, inverse
        ) = self._obfuscator_shortcuts(size, prime, isalt)
        inner_guid = lambda: (
            _int(generate_raw_guid(guid_size) + node + counter(), BIG)
        )
        self._key = lambda: (
            xsalt ^ ((isalt * inner_guid() + osalt) % prime)
        ).to_bytes(size, BIG)
        self._unmask = lambda guid: (
            offset_npad
            ^ (inverse * ((xsalt ^ _int(guid, BIG)) - osalt) % prime)
        ).to_bytes(size, BIG)

    async def anew(
        self,
        *,
        encode: bool = False,
        encoder: t.Callable[..., bytes] = BytesIO.abytes_to_urlsafe,
    ) -> bytes:
        r"""
        Produces a raw-bytes pseudo-random identifier that is guaranteed
        to be unique if each call occurs on a different nanosecond & is
        produced using the same instance `salt`. Additionally, any calls
        which utilize the same `salt` but a different `node_number` will
        always produce unique outputs from each other, even if they
        occur on the same nanosecond. Also, outputs are guaranteed to be
        unique for all calls from any unique instance that utilizes the
        same `salt` & `node_number` if no more than 256 calls are made
        every nanosecond. The probability of per-nanosecond uniqueness
        for instances using the same `salt` increases exponentially with
        linear increases to the instance `size` parameter.

        Normal birthday-bound collision probabilities apply when salts
        are not the same between callers.
        --------
        WARNING: The produced identifiers are randomized & obfuscated
        -------- but they're also invertible. The user must beware not
        to expose the guids if the `node_number`, `salt` or the current
        time in nanoseconds must remain secret from the audiences able
        to view the guids.

        If ``encode`` is set to `True`, then the result is first passed
        as an argument to the ``encoder`` async callable before being
        returned. By default, the ``encoder`` transforms the raw bytes
        into a url safe base64 value without b"=" characters.

         _____________________________________
        |                                     |
        |     Usage Example: Primary Keys     |
        |_____________________________________|

        import aiootp

        dds = distributed_database_system
        shared_salt = dds.shared_salt
        guid = aiootp.GUID(shared_salt, node_number=local_db.id, size=16)

        for table in local_db.tables:
            for record in table.new_records:
                record.dds_primary_key = await guid.anew()
                dds.merge(table, record)

         _____________________________________
        |                                     |
        |     Usage Example: Node Numbers     |
        |_____________________________________|

        import aiootp

        class LargeNetworkGUID(aiootp.GUID, node_number_bytes=4):
            pass

        for i, node in enumerate(network.nodes):  # 256**4 possible nodes
            node.install_guid_generator(
                guid=LargeNetworkGUID(network.salt, node_number=i, size=16)
            )

         _____________________________________
        |                                     |
        |  Usage Example: Configuration Demo  |
        |_____________________________________|

        import aiootp
        from io import StringIO

        # 16 raw bytes without encoding ->
        guid = aiootp.GUID(size=16)
        print(repr(guid.new()))
        b'x]\xd2\xe6\x0c\xb2F\xf9\x05\x02\xa9\xf1\x84\xa3\x0c&'

        # urlsafe base64 encoding ->
        print(repr(guid.new(encode=True)))
        b'3umIg-hZ2J_g8malraZpnw'

        # custom encoding ->
        TRADITIONAL_LAYOUT = (8, 4, 4, 4, 12)

        async def hex_segment_guid(guid: bytes) -> str:
            await asyncio.sleep(0)
            read = StringIO(guid.hex()).read
            return "-".join(read(size) for size in TRADITIONAL_LAYOUT)

        print(repr(await guid.anew(encode=True, encoder=hex_segment_guid)))
        '5e152c91-2f27-8ac6-fa4c-4041ba23d93d'
        """
        if encode:
            return await encoder(self._key())
        await asleep()
        return self._key()

    def new(
        self,
        *,
        encode: bool = False,
        encoder: t.Callable[..., bytes] = BytesIO.bytes_to_urlsafe,
    ) -> bytes:
        r"""
        Produces a raw-bytes pseudo-random identifier that is guaranteed
        to be unique if each call occurs on a different nanosecond & is
        produced using the same instance `salt`. Additionally, any calls
        which utilize the same `salt` but a different `node_number` will
        always produce unique outputs from each other, even if they
        occur on the same nanosecond. Also, outputs are guaranteed to be
        unique for all calls from any unique instance that utilizes the
        same `salt` & `node_number` if no more than 256 calls are made
        every nanosecond. The probability of per-nanosecond uniqueness
        for instances using the same `salt` increases exponentially with
        linear increases to the instance `size` parameter.

        Normal birthday-bound collision probabilities apply when salts
        are not the same between callers.
        --------
        WARNING: The produced identifiers are randomized & obfuscated
        -------- but they're also invertible. The user must beware not
        to expose the guids if the `node_number`, `salt` or the current
        time in nanoseconds must remain secret from the audiences able
        to view the guids.

        If ``encode`` is set to `True`, then the result is first passed
        as an argument to the ``encoder`` async callable before being
        returned. By default, the ``encoder`` transforms the raw bytes
        into a url safe base64 value without b"=" characters.

         _____________________________________
        |                                     |
        |     Usage Example: Primary Keys     |
        |_____________________________________|

        import aiootp

        dds = distributed_database_system
        shared_salt = dds.shared_salt
        guid = aiootp.GUID(shared_salt, node_number=local_db.id, size=16)

        for table in local_db.tables:
            for record in table.new_records:
                record.dds_primary_key = guid.new()
                dds.merge(table, record)

         _____________________________________
        |                                     |
        |  Usage Example: Configuration Demo  |
        |_____________________________________|

        import aiootp
        from io import StringIO

        # 16 raw bytes without encoding ->
        guid = aiootp.GUID(size=16)
        print(repr(guid.new()))
        b'x]\xd2\xe6\x0c\xb2F\xf9\x05\x02\xa9\xf1\x84\xa3\x0c&'

        # urlsafe base64 encoding ->
        print(repr(guid.new(encode=True)))
        b'3umIg-hZ2J_g8malraZpnw'

        # custom encoding ->
        TRADITIONAL_LAYOUT = (8, 4, 4, 4, 12)

        def hex_segment_guid(guid: bytes) -> str:
            read = StringIO(guid.hex()).read
            return "-".join(read(size) for size in TRADITIONAL_LAYOUT)

        print(repr(guid.new(encode=True, encoder=hex_segment_guid)))
        '5e152c91-2f27-8ac6-fa4c-4041ba23d93d'
        """
        if encode:
            return encoder(self._key())
        return self._key()

    async def aunmask(
        self,
        guid: bytes,
        *,
        decode: bool = False,
        decoder: t.Callable[..., bytes] = BytesIO.aurlsafe_to_bytes,
    ) -> UnmaskedGUID:
        """
        Unmasks & optionally, if encoded, decodes, a guid generated by
        an instance which utilized the same instance `salt`, returned in
        an object where the `node_number`, nanosecond `timestamp`, 1-
        byte `counter` & ephemeral `entropy` are accessible via dotted
        attribute lookup. The returned objects are also sortable, which
        sorts first according to the nanosecond `timestamp`, then by the
        instance `node_number`, & finally by the `counter` value.
        """
        if decode:
            guid = await decoder(guid)
        return UnmaskedGUID(self._unmask(guid), self._NODE_NUMBER_BYTES)

    def unmask(
        self,
        guid: bytes,
        *,
        decode: bool = False,
        decoder: t.Callable[..., bytes] = BytesIO.urlsafe_to_bytes,
    ) -> UnmaskedGUID:
        """
        Unmasks & optionally, if encoded, decodes, a guid generated by
        an instance which utilized the same instance `salt`, returned in
        an object where the `node_number`, nanosecond `timestamp`, 1-
        byte `counter` & ephemeral `entropy` are accessible via dotted
        attribute lookup. The returned objects are also sortable, which
        sorts first according to the nanosecond `timestamp`, then by the
        instance `node_number`, & finally by the `counter` value.
        """
        if decode:
            guid = decoder(guid)
        return UnmaskedGUID(self._unmask(guid), self._NODE_NUMBER_BYTES)


async def agenerate_key(
    size: int = KEY_BYTES, *, freshness: int = 8
) -> bytes:
    """
    Returns a random ``size``-byte cryptographically secure key >= 64
    bytes. Defaults to 168-bytes. More entropy can be gathered for
    creating the key using the ``freshness`` parameter. It tells the
    package's chaotic & cpu-intensive `random_number_generator` how many
    concurrent internal asynchronous tasks to sprout for the process of
    gathering entropy.
    """
    if size < MIN_KEY_BYTES or size.__class__ is not int:
        raise Issue.invalid_value("key size", "<64")
    return await arandom_number_generator(
        size, entropy=csprng(), freshness=freshness
    )


def generate_key(size: int = KEY_BYTES, *, freshness: int = 8) -> bytes:
    """
    Returns a random ``size``-byte cryptographically secure key >= 64
    bytes. Defaults to 168-bytes. More entropy can be gathered for
    creating the key using the ``freshness`` parameter. It tells the
    package's chaotic & cpu-intensive `random_number_generator` how many
    concurrent internal asynchronous tasks to sprout for the process of
    gathering entropy.
    """
    if size < MIN_KEY_BYTES or size.__class__ is not int:
        raise Issue.invalid_value("key size", "<64")
    return random_number_generator(
        size, entropy=csprng(), freshness=freshness
    )


async def acsprng(
    entropy: t.Any = run(arandom_number_generator(freshness=1))
) -> bytes:
    """
    Takes in an arbitrary ``entropy`` value from the user to seed then
    return a 64-byte cryptographically secure pseudo-random value.
    """
    if not entropy:
        entropy = _pool[0]
    elif entropy.__class__ is not bytes:
        entropy = repr(entropy).encode() + _pool[0]
    token = await atoken_bytes(32)
    output = await _entropy.ahash(token, entropy)
    thread_safe_entropy = _entropy.copy()
    return await thread_safe_entropy.ahash(token, entropy, output)


def csprng(entropy: t.Any = random_number_generator(freshness=1)) -> bytes:
    """
    Takes in an arbitrary ``entropy`` value from the user to seed then
    return a 64-byte cryptographically secure pseudo-random value.
    """
    if not entropy:
        entropy = _pool[0]
    elif entropy.__class__ is not bytes:
        entropy = repr(entropy).encode() + _pool[0]
    token = token_bytes(32)
    output = _entropy.hash(token, entropy)
    thread_safe_entropy = _entropy.copy()
    return thread_safe_entropy.hash(token, entropy, output)


extras = dict(
    GUID=GUID,
    SequenceID=SequenceID,
    _EntropyDaemon=EntropyDaemon,
    __doc__=__doc__,
    __package__=__package__,
    _agenerate_raw_guid=agenerate_raw_guid,
    _generate_raw_guid=generate_raw_guid,
    achoice=achoice,
    acsprng=acsprng,
    agenerate_salt=agenerate_salt,
    arandom_number_generator=arandom_number_generator,
    arandom_sleep=arandom_sleep,
    atoken_bits=atoken_bits,
    atoken_bytes=atoken_bytes,
    auniform=auniform,
    aunique_range=aunique_range,
    choice=choice,
    csprng=csprng,
    generate_salt=generate_salt,
    random_number_generator=random_number_generator,
    random_sleep=random_sleep,
    token_bits=token_bits,
    token_bytes=token_bytes,
    uniform=uniform,
    unique_range=unique_range,
)


randoms = make_module("randoms", mapping=extras)




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


from test_initialization import *


def test_sign_and_verify():
    try:
        test_key = token_bytes(64)
        test_salt = token_bytes(64)
        test_directory = _paths.DatabasePath()

        signer = PackageSigner(
            package=aiootp.__package__,
            version=aiootp.__version__,
            author=aiootp.__author__,
            license=aiootp.__license__,
            description=aiootp.__doc__,
        )
        signer.update_scope(build_number=0)
        _repr = repr(signer)
        assert all(name in _repr for name in signer._scope)

        context = "Allowed to retrieve a signature before connecting to the database."
        with ignore(RuntimeError, if_else=violation(context)):
            signer._signature

        signer.connect_to_secure_database(
            username=b"test_username",
            passphrase=test_key,
            salt=test_salt,
            path=test_directory,
            mb=MIN_MB,
            cpu=2,
        )
        signer.update_public_credentials(x25519_public_key=aiootp.__PUBLIC_X25519_KEY__)

        context = "Allowed to retrieve a signing key before its creation."
        with ignore(LookupError, if_else=violation(context)):
            signer.signing_key

        context = "Allowed to retrieve a signature before signing."
        with ignore(RuntimeError, if_else=violation(context)):
            signer._signature

        test_signing_key = signer.generate_signing_key()

        # update_signing_key expects a str, bytes, Ed25519PrivateKey or an Ed25519 object
        signer.update_signing_key(test_signing_key)
        signer.update_signing_key(signer.signing_key.secret_key)
        signer.update_signing_key(signer.signing_key.secret_bytes)
        signer.update_signing_key(signer.signing_key.secret_bytes.hex())

        context = "a type other than str, bytes, Ed25519PrivateKey or an Ed25519 object was allowed for update_signing_key"
        with ignore(TypeError, if_else=violation(context)):
            signer.update_signing_key(bytearray(signer.signing_key.secret_bytes))

        filename_sheet = """
        include tests/test_initialization.py
        include tests/test_aiootp.py
        include tests/test_generics.py
        include tests/test_BytesIO.py
        include tests/test_Comprende.py
        include tests/test_misc_in_generics.py
        include tests/test_randoms.py
        include tests/test_misc_in_randoms.py
        include tests/test_ciphers.py
        include tests/test_Database_AsyncDatabase.py
        include tests/test_StreamHMAC.py
        include tests/test_X25519_Ed25519.py
        include tests/test_high_level_encryption.py
        include tests/test_misc_in_ciphers.py
        include tests/test_passcrypt_apasscrypt.py
        """.strip().split("\n")

        test_path = _paths.Path(__file__).parent.parent
        for line in filename_sheet:
            filename = line.strip().split(" ")[-1]
            with open(test_path / filename, "rb") as source_file:
                file_data = source_file.read()
                signer.add_file(filename, file_data)

                # added file hashes are available from the files attribute
                # by their filename
                assert filename in signer.files.namespace
                assert signer.files[filename].digest() != signer._Hasher().digest()
                assert signer.files[filename].digest() == signer._Hasher(file_data).digest()

        signer.sign_package()
        signer.db.save_database()
        summary = signer.summarize()

        # The package verifier successfully verifies a correct summary
        verifier = PackageVerifier(signer.signing_key.public_bytes, path=test_path)
        verifier.verify_summary(summary)

        # Package verifier can also accept an Ed25519 object
        verifier = PackageVerifier(
            Ed25519().import_public_key(signer.signing_key.public_key), path=test_path
        )
        verifier.verify_summary(summary)

        # summarize calls produce the same summaries if the signer object
        # has not been modified
        assert summary == signer.summarize()

        # summary checksum alteration fails
        summary["checksum"] = summary["checksum"][::-1]
        context = "Summary alteration uncaught!"
        with ignore(ValueError, if_else=violation(context)):
            verifier.verify_summary(summary)

        # returning the checksum to the original value succeeds
        summary["checksum"] = summary["checksum"][::-1]
        verifier.verify_summary(summary)

        # a json summary also works
        verifier.verify_summary(json.dumps(summary))

        # altering the signature does not work
        context = "Signature alteration went uncaught!"
        with ignore(verifier.InvalidSignature, if_else=violation(context)):
            fake_summary = {**summary, "signature": token_bytes(64).hex()}
            verifier.verify_summary(fake_summary)

        # the signing key in the summary is the hex representation of the
        # signing objects public key
        assert len(summary["signing_key"]) == 64
        assert summary["signing_key"] == signer.signing_key.public_bytes.hex()

        # altering the signing key fails
        summary["signing_key"] = X25519().generate().public_bytes.hex()
        context = "Changed signing_key went uncaught!"
        with ignore(ValueError, if_else=violation(context)):
            verifier.verify_summary(summary)

        # returning to the original signing key succeeds
        summary["signing_key"] = test_signing_key.public_bytes.hex()
        verifier.verify_summary(summary)

        # changing the signature is detected during summarization
        VERSION = signer._scope.version
        VERSIONS = signer._VERSIONS
        PACKAGE = signer._scope.package
        signature = signer.db[PACKAGE][VERSIONS][VERSION]
        signer.db[PACKAGE][VERSIONS][VERSION] = token_bytes(32).hex()
        context = "altered signature not detected during summarization!"
        with ignore(ValueError, if_else=violation(context)):
            signer.summarize()

        # returning the signature works
        signer.db[PACKAGE][VERSIONS][VERSION] = signature
        signer.summarize()
    finally:
        signer.db.delete_database()


__all__ = sorted({n for n in globals() if n.lower().startswith("test")})




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


import sys
import hmac
import json
import pytest
import base64
import builtins
from math import ceil
from pathlib import Path
from functools import partial
from collections import deque
from secrets import token_bytes, randbits
from hashlib import sha3_256, sha3_512, shake_128, shake_256


_PACKAGE_PATH = str(Path(__file__).absolute().parent.parent)
sys.path.insert(0, _PACKAGE_PATH)


_original_globals = globals().copy()
_globals = globals()


import aiootp
_globals.update(aiootp.__dict__)
_globals.update(_debuggers.__dict__)
_globals.update(_exceptions.__dict__)
_globals.update(_containers.__dict__)
_globals.update(_paths.__dict__)
_globals.update(_typing.__dict__)
_globals.update(commons.__dict__)
_globals.update(constants.__dict__)
_globals.update(misc.__dict__)
_globals.update(datasets.__dict__)
_globals.update(passcrypt.__dict__)
_globals.update(chunky2048.__dict__)
_globals.update(asynchs.__dict__)
_globals.update(randoms.__dict__)
_globals.update(gentools.__dict__)
_globals.update(generics.__dict__)
_globals.update(ciphers.__dict__)
_globals.update(keygens.__dict__)
_globals.update(_original_globals)
_globals.update(builtins.__dict__)


# Use enable_debugging method to run tests in debug mode. May cause noticable slowdown.
DebugControl.disable_debugging()


# acknowledge private variables
Padding = _Padding
StreamHMAC = _StreamHMAC
SyntheticIV = _SyntheticIV
abytes_decipher = _abytes_decipher
bytes_decipher = _bytes_decipher
abytes_encipher = _abytes_encipher
bytes_encipher = _bytes_encipher
aplaintext_stream = _aplaintext_stream
plaintext_stream = _plaintext_stream
KeyAADBundle = _KeyAADBundle


# create static values for this test session
violation = lambda context: partial(raise_exception, AssertionError(context))
aviolation = lambda context: partial(araise_exception, AssertionError(context))

_entropy = csprng() + token_bytes(32)

key = keygens.generate_key()
salt = randoms.generate_salt()
aad = sha3_256(key + salt).digest()
akey_bundle = run(KeyAADBundle(key=key, salt=salt, aad=aad, allow_dangerous_determinism=True).async_mode())
key_bundle = KeyAADBundle(key=key, salt=salt, aad=aad, allow_dangerous_determinism=True).sync_mode()
cipher = Chunky2048(key)

passphrase_0 = csprng()
passphrase_1 = randoms.token_bytes(32)
passphrases = [passphrase_0, passphrase_1]

salt_0 = csprng()
salt_1 = randoms.token_bytes(32)
salts = [salt_0, salt_1]

passcrypt_settings = OpenNamespace(mb=1, cpu=1, cores=1)

tag = "testing"
atag = "a" + tag
metatag = "clients"
ametatag = "a" + metatag

username = b"test suite"
passphrase = b"terrible low entropy passphrase"
PROFILE = dict(username=username, passphrase=passphrase, salt=salt, aad=aad)
LOW_PASSCRYPT_SETTINGS = OpenNamespace(mb=1, cpu=1, cores=1)
PROFILE_AND_SETTINGS = {**PROFILE, **LOW_PASSCRYPT_SETTINGS}
passcrypt_test_vector_0 = OpenNamespace(
    mb=2,
    cpu=2,
    cores=2,
    tag_size=257,
    salt_size=256,
    aad=b"",
    timestamp=bytes.fromhex("000a59139c603129"),
    passphrase=b"test vector passphrase I",
    salt=bytes.fromhex(
        "b3c100b7670ef0c9e15bf877a80d78342ab1f6039511d8444ef82ee9f185fd"
        "ff3bebf168f4af41c5d735a2a6a19c8aab2d142353e307968d9d65e33c9424"
        "a51e0b70c4415723d5c72267cd992fd51141c2ce6ae94ccff9945d1b222fc0"
        "0be81504032ab3e51d0ed617d51fee6cfd5c5302bbd7ada60e0f3d45d631f6"
        "a7c4dc47517f20659a1c00f0f371d8a2e09bb247d4493502b4460f3826a094"
        "ac50faa39b8473629ae782d7fb3bb6c69dfaf68a52f06bc3f92d5c88d99a1f"
        "064d377dd894ee14993cf4d72cd78a4401d519a78fca5b2d7b4b6557e75507"
        "ab94021f2fcb29a10d87c0deca93607e339c1dd67d2db939ee926028712073"
        "bff2135410a51bb2"
    ),
    tag=bytes.fromhex(
        "3a6ed13d7fa4717777bf1de4ce6a5f17e772f708892e9595dfef5db9556322"
        "c5bd65f29d9430c43997b5206b82d148cb4b4df5e0b255d3e4b9f60e36784a"
        "292096816378ddcf42b014dcc55647c61644747ef146e9f3ecaee403f89f09"
        "1b092cccd57691bac847685730f4ad3d5e13f18420c3aeee07519990c3b0ac"
        "65331c6cd276ba34b76a463ab3e6d8c1a4289913b4c972d6a923612515520e"
        "bb577b4b74664c71e61057c350d1a2d87eda99c1a37b5ba23bc6901355910d"
        "711415dc1b7538e3e2fc9086f6b5cd142b345574c73d262f4fba75eb5a1c18"
        "f32ec8e51ba2eef5f86551b4aad15a32569aaa00b80369cb13e4711e22ce95"
        "aa4010f00bccdd15b6"
    ),
    hash_passphrase_result=bytes.fromhex(
        "000a59139c6031290000010101ffb3c100b7670ef0c9e15bf877a80d78342a"
        "b1f6039511d8444ef82ee9f185fdff3bebf168f4af41c5d735a2a6a19c8aab"
        "2d142353e307968d9d65e33c9424a51e0b70c4415723d5c72267cd992fd511"
        "41c2ce6ae94ccff9945d1b222fc00be81504032ab3e51d0ed617d51fee6cfd"
        "5c5302bbd7ada60e0f3d45d631f6a7c4dc47517f20659a1c00f0f371d8a2e0"
        "9bb247d4493502b4460f3826a094ac50faa39b8473629ae782d7fb3bb6c69d"
        "faf68a52f06bc3f92d5c88d99a1f064d377dd894ee14993cf4d72cd78a4401"
        "d519a78fca5b2d7b4b6557e75507ab94021f2fcb29a10d87c0deca93607e33"
        "9c1dd67d2db939ee926028712073bff2135410a51bb23a6ed13d7fa4717777"
        "bf1de4ce6a5f17e772f708892e9595dfef5db9556322c5bd65f29d9430c439"
        "97b5206b82d148cb4b4df5e0b255d3e4b9f60e36784a292096816378ddcf42"
        "b014dcc55647c61644747ef146e9f3ecaee403f89f091b092cccd57691bac8"
        "47685730f4ad3d5e13f18420c3aeee07519990c3b0ac65331c6cd276ba34b7"
        "6a463ab3e6d8c1a4289913b4c972d6a923612515520ebb577b4b74664c71e6"
        "1057c350d1a2d87eda99c1a37b5ba23bc6901355910d711415dc1b7538e3e2"
        "fc9086f6b5cd142b345574c73d262f4fba75eb5a1c18f32ec8e51ba2eef5f8"
        "6551b4aad15a32569aaa00b80369cb13e4711e22ce95aa4010f00bccdd15b6"
    ),
)

passcrypt_test_vector_1 = OpenNamespace(
    mb=3,
    cpu=3,
    cores=3,
    tag_size=32,
    salt_size=10,
    aad=b"testvector",
    timestamp=bytes.fromhex("000a59328da314c9"),
    passphrase=b"test vector passphrase II",
    salt=bytes.fromhex("ca926bc906fa14b886eb"),
    tag=bytes.fromhex(
        "e93c54b0ab0360a4020b0bb6a3a10bcb45f0359d31f319d43412fbcc801afa"
        "b7"
    ),
    hash_passphrase_result=bytes.fromhex(
        "000a59328da314c9000002020209ca926bc906fa14b886ebe93c54b0ab0360"
        "a4020b0bb6a3a10bcb45f0359d31f319d43412fbcc801afab7"
    ),
)

passcrypt_test_vector_2 = OpenNamespace(
    mb=4,
    cpu=4,
    cores=4,
    tag_size=16,
    salt_size=4,
    aad=b"core_cache_change",
    timestamp=bytes.fromhex("000a59529a517df8"),
    passphrase=b"test vector passphrase III",
    salt=bytes.fromhex("80345361"),
    tag=bytes.fromhex("a677df399db38672de807c9147ec438a"),
    hash_passphrase_result=bytes.fromhex(
        "000a59529a517df800000303030380345361a677df399db38672de807c9147"
        "ec438a"
    ),
)

passcrypt_test_vector_3 = OpenNamespace(
    mb=1,
    cpu=5,
    cores=5,
    tag_size=24,
    salt_size=8,
    aad=bytes.fromhex(
        "000102030405060708090a0b0c0d0e0f101112131415161718191a1b1c1d1e"
        "1f202122232425262728292a2b2c2d2e2f303132333435363738393a3b3c3d"
        "3e3f404142434445464748494a4b4c4d4e4f505152535455565758595a5b5c"
        "5d5e5f606162636465666768696a6b6c6d6e6f707172737475767778797a7b"
        "7c7d7e7f808182838485868788898a8b8c8d8e8f909192939495969798999a"
        "9b9c9d9e9fa0a1a2a3a4a5a6a7a8a9aaabacadaeafb0b1b2b3b4b5b6b7b8b9"
        "babbbcbdbebfc0c1c2c3c4c5c6c7c8c9cacbcccdcecfd0d1d2d3d4d5d6d7d8"
        "d9dadbdcdddedfe0e1e2e3e4e5e6e7e8e9eaebecedeeeff0f1f2f3f4f5f6f7"
        "f8f9fafbfcfdfeff"
    ),
    timestamp=bytes.fromhex("000a59a53cfe5ecd"),
    passphrase=b"test vector passphrase IIII",
    salt=bytes.fromhex("fbd2ee954afb48c3"),
    tag=bytes.fromhex("8e8850245c2b699993930240c9e89fc052cfe0b9e27febf1"),
    hash_passphrase_result=bytes.fromhex(
        "000a59a53cfe5ecd000000040407fbd2ee954afb48c38e8850245c2b699993"
        "930240c9e89fc052cfe0b9e27febf1"
    ),
)

byte_leakage = 16 * b"\x00"
string_leakage = 16 * "0".encode()
plaintext_bytes = b"\xff" +  BLOCKSIZE * b"\x00"
plaintext_string = "f" + BLOCKSIZE * "0"
test_data = {
    "floats": 10000.243,
    "ints": list(range(-16, 16, 8)),
    "dicts": {"testing": {}},
    "lists": 4 * [[]],
    "strings": "testing...",
    "nul": None,
    "bools": [True, False],
}
atest_data = test_data.copy()


# Creating timestamped values early so later testing has to wait less time
test_json_ciphertext = cipher.json_encrypt(test_data, aad=aad)
atest_json_ciphertext = run(cipher.ajson_encrypt(atest_data, aad=aad))
test_token_ciphertext = cipher.make_token(plaintext_bytes, aad=aad)
atest_token_ciphertext = run(cipher.amake_token(plaintext_bytes, aad=aad))

aexpired_passcrypt_hash = run(Passcrypt.ahash_passphrase(passphrase_0, mb=1, cores=1))
expired_passcrypt_hash = Passcrypt.hash_passphrase(passphrase_0, mb=1, cores=1)

clock = Clock(SECONDS)
ns_clock = Clock(NANOSECONDS)
time_start = clock.make_timestamp()


@pytest.fixture(scope="session")
def database():
    print("setup".center(15, "-"))

    db = Database(key=key, preload=True)
    db.save_database()
    yield db

    print("teardown".center(18, "-"))
    db.delete_database()


@pytest.fixture(scope="session")
def async_database():
    print("setup".center(15, "-"))

    db = run(AsyncDatabase(key=key, preload=True))
    yield db

    print("teardown".center(18, "-"))
    run(db.adelete_database())


__all__ = [n for n in globals() if not n.startswith("__")]




# This file is part of aiootp, an asynchronous pseudo one-time pad based
# crypto and anonymity library.
#
# Licensed under the AGPLv3: https://www.gnu.org/licenses/agpl-3.0.html
# Copyright © 2019-2021 Gonzo Investigative Journalism Agency, LLC
#            <gonzo.development@protonmail.ch>
#           © 2019-2023 Richard Machado <rmlibre@riseup.net>
# All rights reserved.
#


__doc__  =  "Aggregated Chunky2048 constants."


from math import ceil
from hashlib import sha3_256, sha3_512, shake_128, shake_256


# Config Constants:
KEY = "key"
KEY_BYTES = 168
MIN_KEY_BYTES = 64

AAD = "aad"
DEFAULT_AAD = b""


INNER_HEADER = "inner_header"

TIMESTAMP = "timestamp"
TIMESTAMP_BYTES = 4  # measures seconds

SIV_KEY = "siv_key"
SIV_KEY_BYTES = 16


HEADER = "header"

SHMAC = "shmac"
SHMAC_BYTES = 32

BLOCK_ID = "block_id"
BLOCK_ID_BYTES = 24
MAX_BLOCK_ID_BYTES = 32
MIN_BLOCK_ID_BYTES = 16
BLOCK_ID_PAD = b"\xb8"

SALT = "salt"
SALT_BYTES = 16

IV = "iv"
IV_BYTES = 16


SEED_KDF = "seed_kdf"
SEED_KDF_TYPE = type(shake_128())
SEED_PAD = b"\xac"

LEFT_KDF = "left_kdf"
LEFT_KDF_TYPE = type(shake_128())
LEFT_PAD = b"\x5c"

RIGHT_KDF = "right_kdf"
RIGHT_KDF_TYPE = type(shake_128())
RIGHT_PAD = b"\x36"

SHMAC_TYPE = type(shake_128())
SHMAC_PAD = b"\x9a"


CHUNKY_2048 = "Chunky2048"
DECRYPTION = "decryption"
ENCRYPTION = "encryption"
PLAINTEXT = "plaintext"
CIPHERTEXT = "ciphertext"
BLOCKSIZE = 256
MIN_PADDING_BLOCKS = 0


# Dynamically Generated:
KEY_NIBBLES = 2 * KEY_BYTES
MIN_KEY_NIBBLES = 2 * MIN_KEY_BYTES


INNER_BODY_SLICE = slice(TIMESTAMP_BYTES + SIV_KEY_BYTES, None)
INNER_HEADER_BYTES = TIMESTAMP_BYTES + SIV_KEY_BYTES
INNER_HEADER_NIBBLES = 2 * INNER_HEADER_BYTES
INNER_HEADER_SLICE = slice(None, INNER_HEADER_BYTES)

TIMESTAMP_NIBBLES = 2 * TIMESTAMP_BYTES
TIMESTAMP_SLICE = slice(TIMESTAMP_BYTES)

SIV_KEY_NIBBLES = 2 * SIV_KEY_BYTES
SIV_KEY_SLICE = slice(TIMESTAMP_BYTES, INNER_HEADER_BYTES)


HEADER_BYTES = SHMAC_BYTES + SALT_BYTES + IV_BYTES
HEADER_NIBBLES = 2 * HEADER_BYTES
HEADER_SLICE = slice(None, HEADER_BYTES)

SHMAC_NIBBLES = 2 * SHMAC_BYTES
SHMAC_SLICE = slice(SHMAC_BYTES)

BLOCK_ID_NIBBLES = 2 * BLOCK_ID_BYTES
BLOCK_ID_SLICE = slice(None, BLOCK_ID_BYTES)
MAX_BLOCK_ID_NIBBLES = 2 * MAX_BLOCK_ID_BYTES
MIN_BLOCK_ID_NIBBLES = 2 * MIN_BLOCK_ID_BYTES

SALT_NIBBLES = 2 * SALT_BYTES
SALT_SLICE = slice(SHMAC_BYTES, SHMAC_BYTES + SALT_BYTES)

IV_NIBBLES = 2 * IV_BYTES
IV_SLICE = slice(SHMAC_BYTES + SALT_BYTES, HEADER_BYTES)


SEED_KDF_BLOCKSIZE = shake_128().block_size
SEED_KDF_DOUBLE_BLOCKSIZE = 2 * SEED_KDF_BLOCKSIZE
SEED_KDF_OFFSET = (SEED_KDF_BLOCKSIZE // 2) * SEED_PAD
SEED_RATCHET_KEY_SLICE = slice(SEED_KDF_BLOCKSIZE, None)

LEFT_KDF_BLOCKSIZE = shake_128().block_size
LEFT_RATCHET_KEY_SLICE = slice(0, None, 2)  # Even index bytes

RIGHT_KDF_BLOCKSIZE = shake_128().block_size
RIGHT_RATCHET_KEY_SLICE = slice(1, None, 2)  # Odd index bytes

SHMAC_BLOCKSIZE = shake_128().block_size
SHMAC_DOUBLE_BLOCKSIZE = 2 * SHMAC_BLOCKSIZE


CIPHERTEXT_SLICE = slice(HEADER_BYTES, None)
HALF_BLOCKSIZE = BLOCKSIZE // 2
PACKETSIZE = BLOCK_ID_BYTES + BLOCKSIZE
SENTINEL_BYTES_PER_BLOCKSIZE = ceil(BLOCKSIZE / 256)
PADDING_SENTINEL_BYTES = MIN_PADDING_BLOCKS + SENTINEL_BYTES_PER_BLOCKSIZE
MIN_STREAM_QUEUE = MIN_PADDING_BLOCKS + 1


if INNER_HEADER_BYTES % 2:
    # `SyntheticIV` calculations require the len(inner_header) to be
    # even, otherwise plaintext bytes could be leaked when salt reuse
    # / misuse resistance fails.
    raise ValueError("INNER_HEADER_BYTES *cannot* be an odd number!")
elif BLOCKSIZE % 2:
    # The keystream consists of a left & right KDF which produce equal
    # length keys equal to half the blocksize. An odd number blocksize
    # could leak plaintext as the keys may not sum to the blocksize.
    # It also introduces an unnecessary bias.
    raise ValueError("BLOCKSIZE *cannot* be an odd number!")
elif BLOCKSIZE <= 0:
    # A negative or zero blocksize can not make sense.
    raise ValueError("BLOCKSIZE *cannot* be 0 or negative!")
elif LEFT_KDF_BLOCKSIZE != RIGHT_KDF_BLOCKSIZE:
    # The output KDFs have to produce equal digest outputs.
    raise ValueError("left & right blocksizes *must* be equal!")
elif BLOCKSIZE > (LEFT_KDF_BLOCKSIZE + RIGHT_KDF_BLOCKSIZE):
    # The blocksize of the output KDFs is intimately connected with the
    # arguments of secuirty of the cipher. Particularly, exposing an
    # amount of bytes larger than their internal blocksize, to XOR with
    # plaintext, can only have negative impacts on security.
    raise ValueError(
        "BLOCKSIZE *must* sum to <= the left + right blocksizes!"
    )
elif BLOCKSIZE < INNER_HEADER_BYTES:
    # The inner-header must fit in the first block, therefore the
    # blocksize of the cipher cannot be smaller than it.
    raise ValueError(
        "BLOCKSIZE *must* be at least the size of the INNER_HEADER!"
    )
elif BLOCKSIZE < MIN_KEY_BYTES:
    # The cipher produces a stream of keys which are the length of the
    # blocksize, therefore it cannot be smaller than the minimum allowed
    # size for a key.
    raise ValueError(
        "BLOCKSIZE *must* be at least the size of the MIN_KEY_BYTES!"
    )


# Create a binding to the module's name for UI
chunky2048 = {n: v for n, v in globals().items() if n[0].isupper()}
chunky2048["__all__"] = [*chunky2048]

